import io
import mmap

import numpy as np
import pytest

from asdf import constants, generic_io
from asdf._block import io as bio
from asdf._block.exceptions import BlockIndexError


def test_checksum(tmp_path):
    my_array = np.arange(0, 64, dtype="<i8")
    target_checksum = b"\xcaM\\\xb8t_L|\x00\n+\x01\xf1\xcfP1"
    assert bio.calculate_block_checksum(my_array) == target_checksum
    assert bio.calculate_block_checksum(my_array.reshape((8, 8))) == target_checksum

    # check that when written, a block generates the correct checksum
    path = tmp_path / "test"
    with generic_io.get_file(path, mode="w") as fd:
        # check that no warnings are raised when writing checksums without compression
        bio.write_block(fd, my_array.view(dtype="uint8"), write_checksum=True)
    with generic_io.get_file(path, mode="r") as fd:
        _, header, _, _ = bio.read_block(fd, True)
    assert header["checksum"] == target_checksum


def test_validate_block_header():
    # check for invalid compression
    with pytest.raises(ValueError):
        bio.validate_block_header({"compression": b"foo"})

    # streamed blocks cannot be compressed
    with pytest.raises(ValueError):
        bio.validate_block_header({"compression": b"zlib", "flags": constants.BLOCK_FLAG_STREAMED})
    bio.validate_block_header(
        {"compression": b"\0\0\0\0", "flags": constants.BLOCK_FLAG_STREAMED, "used_size": 0, "data_size": 0},
    )

    # if not compressed, used_size must equal data_size
    with pytest.raises(ValueError):
        bio.validate_block_header({"compression": b"\0\0\0\0", "flags": 0, "used_size": 1, "data_size": 0})
    bio.validate_block_header({"compression": b"zlib", "flags": 0, "used_size": 1, "data_size": 0})


def test_read_block_header(tmp_path):
    # first write out a file with a block
    my_array = np.arange(0, 64, dtype="<i8")
    path = tmp_path / "test"
    with generic_io.get_file(path, mode="w") as fd:
        written_header = bio.write_block(fd, my_array.view("uint8"))

    # try reading the block header at the start
    with generic_io.get_file(path, mode="r") as fd:
        header = bio.read_block_header(fd)
    assert header == written_header

    # seek into the header, creating an invalid starting position
    # where the header size will be 0
    with generic_io.get_file(path, mode="r") as fd:
        fd.seek(10)
        with pytest.raises(ValueError):
            header = bio.read_block_header(fd)

    # check that offset, when provided is used
    with generic_io.get_file(path, mode="r") as fd, pytest.raises(ValueError):
        header = bio.read_block_header(fd, offset=10)
    with generic_io.get_file(path, mode="r") as fd:
        fd.seek(4)
        header = bio.read_block_header(fd, offset=0)
    assert header == written_header


def test_read_block_data(tmp_path):
    # first write out a file with a block
    my_array = np.arange(0, 64, dtype="uint8")
    path = tmp_path / "test"
    with generic_io.get_file(path, mode="w") as fd:
        bio.write_block(fd, my_array)

    # read the header and get the offset within the file where the
    # data is located
    with generic_io.get_file(path, mode="r") as fd:
        header = bio.read_block_header(fd)
        data_offset = fd.tell()
        data = bio.read_block_data(fd, header, True)
        assert np.array_equal(data, my_array)
        # check that offset works as expected
        data = bio.read_block_data(fd, header, True, offset=data_offset)
        assert np.array_equal(data, my_array)

    # not lazy loaded, memmapped
    with generic_io.get_file(path, mode="r") as fd:
        offset, header, read_data_offset, data = bio.read_block(fd, True, lazy_load=False, memmap=True)
        assert offset == 0
        assert read_data_offset == data_offset
        assert np.array_equal(data, my_array)
        base = data
        while getattr(base, "base", None) is not None:
            base = base.base
        assert isinstance(base, mmap.mmap)

    # not memmapped, lazy
    with generic_io.get_file(path, mode="r") as fd:
        offset, header, read_data_offset, data = bio.read_block(fd, True, lazy_load=True, memmap=False)
        assert offset == 0
        assert read_data_offset == data_offset
        assert np.array_equal(data(), my_array)
        base = data()
        while getattr(base, "base", None) is not None:
            base = base.base
        assert not isinstance(base, mmap.mmap)


def test_read_block(tmp_path):
    block_bytes = io.BytesIO(
        b"\x000"  # header size = 2
        + b"\0\0\0\0"  # flags = 4
        + b"\0\0\0\0"  # compression = 4
        + b"\0\0\0\0\0\0\0\0"  # allocated size = 8
        + b"\0\0\0\0\0\0\0\0"  # used size = 8
        + b"\0\0\0\0\0\0\0\0"  # data size = 8
        + b"\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~",  # checksum = 16
    )

    with generic_io.get_file(block_bytes, mode="r") as fd:
        offset, header, data_offset, data = bio.read_block(fd, True)
        assert offset == 0
        assert data_offset == 50
        assert len(data) == 0
        for k in "flags", "allocated_size", "used_size", "data_size":
            assert header[k] == 0


def test_write_block(tmp_path):
    expected = (
        b"\x000"  # header size = 2
        + b"\0\0\0\0"  # flags
        + b"\0\0\0\0"  # compression
        + b"\0\0\0\0\0\0\0\0"  # allocated size
        + b"\0\0\0\0\0\0\0\0"  # used size
        + b"\0\0\0\0\0\0\0\0"  # data size
        + b"\xd4\x1d\x8c\xd9\x8f\x00\xb2\x04\xe9\x80\t\x98\xec\xf8B~"  # checksum
    )

    # first write out a file with a block
    my_array = np.array([], dtype="uint8")
    path = tmp_path / "test"
    with generic_io.get_file(path, mode="w") as fd:
        bio.write_block(fd, my_array)

    with generic_io.get_file(path, mode="r") as fd:
        buff = fd.read(-1)
        assert buff == expected


def test_write_block_offset():
    data = np.ones(30, dtype="uint8")
    raw_fd = io.BytesIO()
    fd = generic_io.get_file(raw_fd, mode="rw")
    fd.write(b"0000000")
    bio.write_block(fd, data, offset=0)
    fd.seek(0)
    _, _, _, d = bio.read_block(fd, True)
    np.testing.assert_array_equal(data, d)


def test_write_oversized_block():
    # check that writing does not occur if data_size > allocated_size
    data = np.ones(30, dtype="uint8")
    raw_fd = io.BytesIO()
    fd = generic_io.get_file(raw_fd, mode="rw")
    with pytest.raises(RuntimeError, match=r"Block used size.*"):
        bio.write_block(fd, data, allocated_size=0)
    assert fd.tell() == 0


def test_fd_not_seekable():
    data = np.ones(30, dtype="uint8")
    raw_fd = io.BytesIO()
    fd = generic_io.get_file(raw_fd, mode="rw")
    bio.write_block(fd, data)

    raw_fd.seek(0)
    fd = generic_io.get_file(raw_fd, mode="rw")

    seekable = lambda: False  # noqa: E731
    fd.seekable = seekable

    _, _, _, d = bio.read_block(fd, True)

    np.testing.assert_array_equal(d, data)

    with pytest.raises(ValueError, match=r"write_block received offset.*"):
        bio.write_block(fd, data, offset=0)


def test_compressed_block():
    data = np.ones(30, dtype="uint8")
    fd = generic_io.get_file(io.BytesIO(), mode="rw")
    write_header = bio.write_block(fd, data, compression="zlib")
    assert write_header["compression"] == b"zlib"
    _, _, _, rdata = bio.read_block(fd, True, offset=0)
    np.testing.assert_array_equal(rdata, data)


def test_stream_block():
    data = np.ones(10, dtype="uint8")
    fd = generic_io.get_file(io.BytesIO(), mode="rw")
    write_header = bio.write_block(fd, data, stream=True)
    assert write_header["flags"] & constants.BLOCK_FLAG_STREAMED
    # now write extra data to file
    extra_data = np.ones(10, dtype="uint8")
    fd.write_array(extra_data)
    _, _, _, rdata = bio.read_block(fd, True, offset=0)
    assert rdata.size == 20
    assert np.all(rdata == 1)


def test_read_from_closed(tmp_path):
    fn = tmp_path / "test.blk"
    data = np.ones(10, dtype="uint8")
    with generic_io.get_file(fn, mode="w") as fd:
        bio.write_block(fd, data, stream=True)
    with generic_io.get_file(fn, mode="rw") as fd:
        _, _, _, callback = bio.read_block(fd, True, offset=0, lazy_load=True)
    with pytest.raises(OSError, match=r"ASDF file has already been closed\. Can not get the data\."):
        callback()


@pytest.mark.parametrize("data", [np.ones(10, dtype="f4"), np.ones((3, 3), dtype="uint8")])
def test_invalid_data(data):
    fd = generic_io.get_file(io.BytesIO(), mode="rw")
    with pytest.raises(ValueError, match=r"Data must be of.*"):
        bio.write_block(fd, data, stream=True)


@pytest.mark.parametrize(
    "options",
    [
        (0, 10, 5, [5, 0]),
        (0, 10, 3, [9, 6, 3, 0]),
        (0, 10, 10, [0]),
        (0, 10, 6, [6, 0]),
        (0, 10, 11, [0]),
        (0, 10, 4096, [0]),
    ],
)
def test_candidate_offsets(options):
    min_offset, max_offset, size, targets = options
    for offset, target in zip(bio._candidate_offsets(min_offset, max_offset, size), targets):
        assert offset == target


def generate_block_index_file(fn, values=None, offset=0):
    if values is None:
        values = [1, 2, 3]
    with generic_io.get_file(fn, "w") as f:
        f.write(b"\0" * offset)
        bio.write_block_index(f, values)


def test_find_block_index(tmp_path):
    offset = 42
    fn = tmp_path / "test"
    generate_block_index_file(fn, offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.find_block_index(fd) == offset


def test_find_block_index_on_boundry(tmp_path):
    fn = tmp_path / "test"
    with generic_io.get_file(fn, "w") as fd:
        block_size = fd.block_size
    # put pattern across a block boundary
    offset = block_size - (len(constants.INDEX_HEADER) // 2)
    generate_block_index_file(fn, offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.find_block_index(fd) == offset


def test_missing_block_index(tmp_path):
    fn = tmp_path / "test"
    with open(fn, "w") as f:
        f.write("\0" * 4096)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.find_block_index(fd) is None


def test_less_than_min_offset_block_index(tmp_path):
    fn = tmp_path / "test"
    offset = 26
    min_offset = 42
    generate_block_index_file(fn, offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.find_block_index(fd, min_offset) is None


def test_greater_than_max_offset_block_index(tmp_path):
    fn = tmp_path / "test"
    offset = 72
    max_offset = 42
    generate_block_index_file(fn, offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.find_block_index(fd, 0, max_offset) is None


def test_read_block_index(tmp_path):
    fn = tmp_path / "test"
    values = [1, 2, 3]
    generate_block_index_file(fn, values=values, offset=0)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.read_block_index(fd) == values


def test_read_block_index_with_offset(tmp_path):
    fn = tmp_path / "test"
    values = [1, 2, 3]
    offset = 42
    generate_block_index_file(fn, values=values, offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.read_block_index(fd, offset) == values


def test_read_block_index_pre_seek(tmp_path):
    fn = tmp_path / "test"
    values = [1, 2, 3]
    offset = 42
    generate_block_index_file(fn, values=values, offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        fd.seek(offset)
        assert bio.read_block_index(fd) == values


def test_read_block_index_no_header(tmp_path):
    fn = tmp_path / "test"
    values = [1, 2, 3]
    generate_block_index_file(fn, values=values, offset=0)
    with generic_io.get_file(fn, "r") as fd:
        fd.seek(len(constants.INDEX_HEADER))
        with pytest.raises(BlockIndexError, match=r"Failed to read block index.*"):
            assert bio.read_block_index(fd) == values


def test_read_block_index_invalid_yaml():
    bs = io.BytesIO(constants.INDEX_HEADER + b"][")
    with generic_io.get_file(bs, "r") as fd:
        with pytest.raises(BlockIndexError, match="Failed to parse block index as yaml"):
            bio.read_block_index(fd)


def test_read_block_index_valid_yaml_invalid_contents():
    bs = io.BytesIO(constants.INDEX_HEADER + b"['a', 'b']")
    with generic_io.get_file(bs, "r") as fd:
        with pytest.raises(BlockIndexError, match="Invalid block index"):
            bio.read_block_index(fd)


def test_write_block_index_with_offset(tmp_path):
    fn = tmp_path / "test"
    offset = 50
    with generic_io.get_file(fn, "w") as fd:
        fd.write(b"\0" * 100)
        fd.seek(0)
        bio.write_block_index(fd, [1, 2, 3], offset=offset)
    with generic_io.get_file(fn, "r") as fd:
        assert bio.find_block_index(fd) == offset



import argparse
from pathlib import Path

from common import generate_file

from asdf.versioning import AsdfVersion


def parse_args():
    parser = argparse.ArgumentParser(description="Generate an ASDF file for library version compatibility testing")
    parser.add_argument("filename", help="the output filename")
    parser.add_argument("version", help="the ASDF Standard version to write")
    return parser.parse_args()


def main():
    args = parse_args()

    path = Path(args.filename)
    version = AsdfVersion(args.version)

    generate_file(path, version)


if __name__ == "__main__":
    main()



import gc

import numpy as np
import pytest

import asdf
from asdf._block import manager
from asdf._block.options import Options


def test_set_streamed_block_via_options():
    options = manager.OptionsStore(manager.ReadBlocks())
    arr1 = np.arange(10, dtype="uint8")
    arr2 = np.arange(5, dtype="uint8")
    options.set_options(arr1, Options("streamed"))
    with pytest.raises(ValueError, match=r"Can not add second streaming block"):
        options.set_options(arr2, Options("streamed"))
    del arr1
    gc.collect(2)
    options.set_options(arr2, Options("streamed"))


def test_set_streamed_block_via_manager():
    af = asdf.AsdfFile()
    m = af._blocks

    class Foo:
        pass

    arr = np.arange(10, dtype="uint8")
    obj = Foo()
    m.set_streamed_write_block(arr, obj)

    # setting again with the same data is ok
    m.set_streamed_write_block(arr, obj)

    # using a different array is not allowed
    arr2 = np.arange(3, dtype="uint8")
    with pytest.raises(ValueError, match="Can not add second streaming block"):
        m.set_streamed_write_block(arr2, obj)

    # a different object is ok as long as the array matches
    obj2 = Foo()
    m.set_streamed_write_block(arr, obj2)


def test_load_external_internal(tmp_path):
    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"arr": np.arange(10, dtype="uint8")}).write_to(fn)
    with asdf.open(fn) as af:
        m = af._blocks
        np.testing.assert_array_equal(m._load_external("#"), m.blocks[0].data)


def test_write_no_uri(tmp_path):
    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"arr": np.arange(10, dtype="uint8")}).write_to(fn)
    with asdf.open(fn) as af:
        m = af._blocks
        with pytest.raises(ValueError, match=r"Can't write external blocks.*"):
            m._write_external_blocks(write_checksums=False)


def test_write_outside_context(tmp_path):
    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"arr": np.arange(10, dtype="uint8")}).write_to(fn)
    with asdf.open(fn) as af:
        m = af._blocks
        with pytest.raises(OSError, match=r"write called outside of valid write_context"):
            m.write(pad_blocks=False, include_block_index=False, write_checksums=False)


def test_update_outside_context(tmp_path):
    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"arr": np.arange(10, dtype="uint8")}).write_to(fn)
    with asdf.open(fn) as af:
        m = af._blocks
        with pytest.raises(OSError, match=r"update called outside of valid write_context"):
            m.update(0, pad_blocks=False, include_block_index=False, write_checksums=False)


def test_input_compression(tmp_path):
    fn = tmp_path / "test.asdf"
    af = asdf.AsdfFile({"arr": np.arange(10, dtype="uint8")})
    af.set_array_compression(af["arr"], "zlib")
    af.write_to(fn)

    with asdf.open(fn) as af:
        assert af.get_array_compression(af["arr"]) == "zlib"
        af.set_array_compression(af["arr"], "bzp2")
        assert af.get_array_compression(af["arr"]) == "bzp2"
        af.set_array_compression(af["arr"], "input")
        assert af.get_array_compression(af["arr"]) == "zlib"



import numpy as np

import asdf


def test_memmap_view_access_after_close(tmp_path):
    """
    Accessing a view of a memmap after the asdf file
    is closed results in a segfault

    https://github.com/asdf-format/asdf/issues/1334
    """

    a = np.ones(10, dtype="uint8")
    fn = tmp_path / "test.asdf"
    asdf.AsdfFile({"a": a}).write_to(fn)

    with asdf.open(fn, memmap=True) as af:
        v = af["a"][:5]

    assert np.all(v == 1)



import io

import asdf


def test_init(tree, benchmark):
    benchmark(asdf.AsdfFile, tree)


def test_validate(asdf_file, benchmark):
    benchmark(asdf_file.validate)


def test_write_to(asdf_file, benchmark):
    bs = io.BytesIO()
    benchmark(asdf_file.write_to, bs)


def test_open(tree_bytes, benchmark):
    # open doesn't seek to start of file so make a new BytesIO each time
    def asdf_open():
        asdf.open(io.BytesIO(tree_bytes))

    benchmark(asdf_open)


def test_update(tree_bytes, benchmark):
    bs = io.BytesIO(tree_bytes)
    af = asdf.open(bs)
    benchmark(af.update)


def test_dump(tree, benchmark):
    benchmark(asdf.dumps, tree)


def test_load(tree_bytes, benchmark):
    benchmark(asdf.loads, tree_bytes)



"""
Test dump, dumps, load and loads

These are thin wrappers around AsdfFile
so the tests here will be relatively simple.
"""

from asdf import dump, dumps, load, loads
from asdf._tests._helpers import assert_tree_match


def test_roundtrip(tmp_path, tree):
    fn = tmp_path / "test.asdf"
    dump(tree, fn)
    rt = load(fn)
    assert_tree_match(tree, rt)


def test_str_roundtrip(tree):
    rt = loads(dumps(tree))
    assert_tree_match(tree, rt)



import argparse
import logging
import sys

from asdf import util

# This list is ordered in order of average workflow
command_order = ["Explode", "Implode"]


class Command:
    @classmethod
    def setup_arguments(cls, subparsers):
        raise NotImplementedError

    @classmethod
    def run(cls, args):
        raise NotImplementedError


def make_argparser():
    """
    Most of the real work is handled by the subcommands in the
    commands subpackage.
    """

    def help_(args):
        parser.print_help()
        return 0

    parser = argparse.ArgumentParser("asdftool", description="Commandline utilities for managing ASDF files.")

    parser.add_argument("--verbose", "-v", action="store_true", help="Increase verbosity")

    subparsers = parser.add_subparsers(title="subcommands", description="valid subcommands")

    help_parser = subparsers.add_parser("help", help="Display usage information")
    help_parser.set_defaults(func=help_)

    commands = {x.__name__: x for x in util._iter_subclasses(Command)}

    for command in command_order:
        commands[str(command)].setup_arguments(subparsers)
        del commands[command]

    for _, command in sorted(commands.items()):
        command.setup_arguments(subparsers)

    return parser, subparsers


def main_from_args(args):
    parser, _ = make_argparser()

    args = parser.parse_args(args)

    # Only needed for Python 3, apparently, but can't hurt
    if not hasattr(args, "func"):
        parser.print_help()
        return 2

    try:
        result = args.func(args)
    except RuntimeError as e:
        logging.error(str(e))
        return 1
    except OSError as e:
        logging.error(str(e))
        return e.errno

    if result is None:
        result = 0

    return result


def main(args=None):
    if args is None:
        args = sys.argv[1:]
    sys.exit(main_from_args(args))



import abc

from packaging.specifiers import SpecifierSet

from asdf.util import get_class_name

from ._compressor import Compressor
from ._converter import ConverterProxy
from ._tag import TagDefinition
from ._validator import Validator


class Extension(abc.ABC):
    """
    Abstract base class defining an extension to ASDF.

    Implementing classes must provide the `extension_uri`.
    Other properties are optional.
    """

    @classmethod
    def __subclasshook__(cls, class_):
        if cls is Extension:
            return hasattr(class_, "extension_uri")
        return NotImplemented  # pragma: no cover

    @property
    @abc.abstractmethod
    def extension_uri(self):
        """
        Get the URI of the extension to the ASDF Standard implemented
        by this class.  Note that this may not uniquely identify the
        class itself.

        Returns
        -------
        str
        """

    @property
    def legacy_class_names(self):
        """
        Get the set of fully-qualified class names used by older
        versions of this extension.  This allows a new-style
        implementation of an extension to prevent warnings when a
        legacy extension is missing.

        Returns
        -------
        iterable of str
        """
        return set()

    @property
    def asdf_standard_requirement(self):
        """
        Get the ASDF Standard version requirement for this extension.

        Returns
        -------
        str or None
            If str, PEP 440 version specifier.
            If None, support all versions.
        """
        return

    @property
    def converters(self):
        """
        Get the `asdf.extension.Converter` instances for tags
        and Python types supported by this extension.

        Returns
        -------
        iterable of asdf.extension.Converter
        """
        return []

    @property
    def tags(self):
        """
        Get the YAML tags supported by this extension.

        Returns
        -------
        iterable of str or asdf.extension.TagDefinition
        """
        return []

    @property
    def compressors(self):
        """
        Get the `asdf.extension.Compressor` instances for
        compression schemes supported by this extension.

        Returns
        -------
        iterable of asdf.extension.Compressor
        """
        return []

    @property
    def yaml_tag_handles(self):
        """
        Get a dictionary of custom yaml TAG handles defined by the extension.

        The dictionary key indicates the TAG handles to be placed in the YAML header,
        the value defines the string for tag replacement.
        See https://yaml.org/spec/1.1/#tag%20shorthand/

        Example: ``{"!foo!": "tag:nowhere.org:custom/"}``

        Returns
        -------
        dict

        """
        return {}

    @property
    def validators(self):
        """
        Get the `asdf.extension.Validator` instances for additional
        schema properties supported by this extension.

        Returns
        -------
        iterable of asdf.extension.Validator
        """
        return []


class ExtensionProxy(Extension):
    """
    Proxy that wraps an extension, provides default implementations
    of optional methods, and carries additional information on the
    package that provided the extension.
    """

    @classmethod
    def maybe_wrap(cls, delegate):
        if isinstance(delegate, ExtensionProxy):
            return delegate

        return ExtensionProxy(delegate)

    def __init__(self, delegate, package_name=None, package_version=None):
        if not isinstance(delegate, Extension):
            msg = "Extension must implement the Extension interface"
            raise TypeError(msg)

        self._delegate = delegate
        self._package_name = package_name
        self._package_version = package_version

        self._class_name = get_class_name(delegate)

        self._legacy = False

        # Sort these out up-front so that errors are raised when the extension is loaded
        # and not in the middle of the user's session.  The extension will fail to load
        # and a warning will be emitted, but it won't crash the program.

        self._legacy_class_names = set()
        for class_name in getattr(self._delegate, "legacy_class_names", []):
            if isinstance(class_name, str):
                self._legacy_class_names.add(class_name)
            else:
                msg = "Extension property 'legacy_class_names' must contain str values"
                raise TypeError(msg)

        if self._legacy:
            self._legacy_class_names.add(self._class_name)

        value = getattr(self._delegate, "asdf_standard_requirement", None)
        if isinstance(value, str):
            self._asdf_standard_requirement = SpecifierSet(value)
        elif value is None:
            self._asdf_standard_requirement = SpecifierSet()
        else:
            msg = "Extension property 'asdf_standard_requirement' must be str or None"
            raise TypeError(msg)

        self._tags = []
        for tag in getattr(self._delegate, "tags", []):
            if isinstance(tag, str):
                self._tags.append(TagDefinition(tag))
            elif isinstance(tag, TagDefinition):
                self._tags.append(tag)
            else:
                msg = "Extension property 'tags' must contain str or asdf.extension.TagDefinition values"
                raise TypeError(msg)

        self._yaml_tag_handles = getattr(delegate, "yaml_tag_handles", {})

        # Process the converters last, since they expect ExtensionProxy
        # properties to already be available.
        self._converters = [ConverterProxy(c, self) for c in getattr(self._delegate, "converters", [])]

        self._compressors = []
        if hasattr(self._delegate, "compressors"):
            for compressor in self._delegate.compressors:
                if not isinstance(compressor, Compressor):
                    msg = "Extension property 'compressors' must contain instances of asdf.extension.Compressor"
                    raise TypeError(msg)
                self._compressors.append(compressor)

        self._validators = []
        if hasattr(self._delegate, "validators"):
            for validator in self._delegate.validators:
                if not isinstance(validator, Validator):
                    msg = "Extension property 'validators' must contain instances of asdf.extension.Validator"
                    raise TypeError(msg)
                self._validators.append(validator)

    @property
    def extension_uri(self):
        """
        Get the URI of the extension to the ASDF Standard implemented
        by this class.  Note that this may not uniquely identify the
        class itself.

        Returns
        -------
        str or None
        """
        return getattr(self._delegate, "extension_uri", None)

    @property
    def legacy_class_names(self):
        """
        Get the set of fully-qualified class names used by older
        versions of this extension.  This allows a new-style
        implementation of an extension to prevent warnings when a
        legacy extension is missing.

        Returns
        -------
        set of str
        """
        return self._legacy_class_names

    @property
    def asdf_standard_requirement(self):
        """
        Get the extension's ASDF Standard requirement.

        Returns
        -------
        packaging.specifiers.SpecifierSet
        """
        return self._asdf_standard_requirement

    @property
    def converters(self):
        """
        Get the extension's converters.

        Returns
        -------
        list of asdf.extension.Converter
        """
        return self._converters

    @property
    def compressors(self):
        """
        Get the extension's compressors.

        Returns
        -------
        list of asdf.extension.Compressor
        """
        return self._compressors

    @property
    def tags(self):
        """
        Get the YAML tags supported by this extension.

        Returns
        -------
        list of asdf.extension.TagDefinition
        """
        return self._tags

    @property
    def types(self):
        """
        Get the legacy extension's ExtensionType subclasses.

        Returns
        -------
        iterable of asdf.type.ExtensionType
        """
        return getattr(self._delegate, "types", [])

    @property
    def tag_mapping(self):
        """
        Get the legacy extension's tag-to-schema-URI mapping.

        Returns
        -------
        iterable of tuple or callable
        """
        return getattr(self._delegate, "tag_mapping", [])

    @property
    def url_mapping(self):
        """
        Get the legacy extension's schema-URI-to-URL mapping.

        Returns
        -------
        iterable of tuple or callable
        """
        return getattr(self._delegate, "url_mapping", [])

    @property
    def delegate(self):
        """
        Get the wrapped extension instance.

        Returns
        -------
        asdf.extension.Extension
        """
        return self._delegate

    @property
    def package_name(self):
        """
        Get the name of the Python package that provided this extension.

        Returns
        -------
        str or None
            `None` if the extension was added at runtime.
        """
        return self._package_name

    @property
    def package_version(self):
        """
        Get the version of the Python package that provided the extension

        Returns
        -------
        str or None
            `None` if the extension was added at runtime.
        """
        return self._package_version

    @property
    def class_name(self):
        """
        Get the fully qualified class name of the extension.

        Returns
        -------
        str
        """
        return self._class_name

    @property
    def legacy(self):
        """
        False
        """
        return self._legacy

    @property
    def yaml_tag_handles(self):
        """
        Get a dictionary of custom yaml TAG handles defined by the extension.

        The dictionary key indicates the TAG handles to be placed in the YAML header,
        the value defines the string for tag replacement.
        See https://yaml.org/spec/1.1/#tag%20shorthand/

        Example: ``{"!foo!": "tag:nowhere.org:custom/"}``

        Returns
        -------
        dict

        """
        return self._yaml_tag_handles

    @property
    def validators(self):
        """
        Get the `asdf.extension.Validator` instances for additional
        schema properties supported by this extension.

        Returns
        -------
        list of asdf.extension.Validator
        """
        return self._validators

    def __eq__(self, other):
        if isinstance(other, ExtensionProxy):
            return other.delegate is self.delegate

        return False

    def __hash__(self):
        return hash(id(self.delegate))

    def __repr__(self):
        package_description = "(none)" if self.package_name is None else f"{self.package_name}=={self.package_version}"

        uri_description = "(none)" if self.extension_uri is None else self.extension_uri

        return (
            f"<ExtensionProxy URI: {uri_description} class: {self.class_name} "
            f"package: {package_description} legacy: {self.legacy}>"
        )



"""
Commands for validating ASDF files
"""

import asdf

from .main import Command

__all__ = ["validate"]


class Validate(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "validate",
            help="Validates an ASDF file.",
            description=(
                "Validates tree against all tagged schemas (and optionally a custom schema) "
                "and blocks with stored checksums."
            ),
        )

        parser.add_argument("filename", help="path to ASDF file")
        parser.add_argument(
            "--custom-schema",
            type=str,
            help="path or URI of custom schema",
        )
        parser.add_argument(
            "--skip-block-validation",
            default=False,
            action="store_true",
            help="Do not compare blocks against stored checksums.",
        )
        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        validate(args.filename, args.custom_schema, args.skip_block_validation)


def validate(filename, custom_schema, skip_block_validation):
    # if we are skipping checksums we can lazy load, otherwise don't
    with asdf.open(
        filename,
        custom_schema=custom_schema,
        validate_checksums=not skip_block_validation,
        lazy_load=skip_block_validation,
    ) as af:  # noqa: F841
        msg = f"{filename} is valid"
        if custom_schema:
            msg += f", conforms to {custom_schema}"
        if not skip_block_validation:
            msg += ", and block checksums match contents"
        print(msg)



from asdf.tags.core import Constant
from asdf.testing import helpers


def test_constant():
    constant = Constant("gardener")

    result = helpers.roundtrip_object(constant)

    assert result.value == constant.value



import datetime
import sys
from pathlib import Path

if sys.version_info < (3, 11):
    import tomli as tomllib
else:
    import tomllib

# The standard library importlib.metadata returns duplicate entrypoints
# for all python versions up to and including 3.11
# https://github.com/python/importlib_metadata/issues/410#issuecomment-1304258228
# see PR https://github.com/asdf-format/asdf/pull/1260
# see issue https://github.com/asdf-format/asdf/issues/1254
if sys.version_info >= (3, 12):
    from importlib.metadata import distribution
else:
    from importlib_metadata import distribution


# Get configuration information from `pyproject.toml`
with open(Path(__file__).parent.parent / "pyproject.toml", "rb") as configuration_file:
    conf = tomllib.load(configuration_file)

configuration = conf["project"]

# -- Project information ------------------------------------------------------
project = configuration["name"]
author = configuration["authors"][0]["name"]
copyright = f"{datetime.datetime.now().year}, {author}"

release = distribution(configuration["name"]).version
# for example take major/minor
version = ".".join(release.split(".")[:2])

# -- Options for HTML output ---------------------------------------------------
html_title = f"{project} v{release}"

# Output file base name for HTML help builder.
htmlhelp_basename = project + "doc"

# -- Options for LaTeX output --------------------------------------------------
latex_documents = [("index", project + ".tex", project + " Documentation", author, "manual")]

# -- Options for manual page output --------------------------------------------
man_pages = [("index", project.lower(), project + " Documentation", [author], 1)]

# Enable nitpicky mode - which ensures that all references in the docs
# resolve.

nitpicky = True

# ignore a few pyyaml docs links since they don't appear to support intersphinx
nitpick_ignore = [
    ("py:class", "yaml.representer.RepresenterError"),
    ("py:class", "yaml.error.YAMLError"),
]

# Add intersphinx mappings
intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pypa-packaging": ("https://packaging.python.org/en/latest/", None),
    "pytest": ("https://docs.pytest.org/en/latest/", None),
    "python": ("https://docs.python.org/3/", None),
    "semantic_version": ("https://python-semanticversion.readthedocs.io/en/latest/", None),
    "stdatamodels": ("https://stdatamodels.readthedocs.io/en/latest/", None),
}


# Docs are hosted as a "subproject" under the main project's domain: https://www.asdf-format.org/projects
# This requires including links to main project (asdf-website) and the other asdf subprojects
# See https://docs.readthedocs.io/en/stable/guides/intersphinx.html#using-intersphinx
subprojects = {
    # main project
    "asdf-website": ("https://www.asdf-format.org/en/latest", None),
    # other subprojects
    "asdf-standard": ("https://www.asdf-format.org/projects/asdf-standard/en/latest/", None),
    "asdf-coordinates-schemas": ("https://www.asdf-format.org/projects/asdf-coordinates-schemas/en/latest/", None),
    "asdf-transform-schemas": ("https://www.asdf-format.org/projects/asdf-transform-schemas/en/latest/", None),
    "asdf-wcs-schemas": ("https://www.asdf-format.org/projects/asdf-wcs-schemas/en/latest/", None),
}


intersphinx_mapping.update(subprojects)

extensions = [
    # TODO clean these up, do we need them all?
    "sphinx_inline_tabs",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx_asdf",
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinxcontrib.jquery",
    "numpydoc",
    "sphinx_automodapi.automodapi",
    "sphinx_automodapi.smart_resolver",
]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The reST default role (used for this markup: `text`) to use for all
# documents. Set to the "smart" one.
default_role = "obj"

# Don't show summaries of the members in each class along with the
# class' docstring
numpydoc_show_class_members = False

autosummary_generate = True

automodapi_toctreedirnm = "api"

# Class documentation should contain *both* the class docstring and
# the __init__ docstring
autoclass_content = "both"

html_theme = "furo"
html_static_path = ["_static"]
html_sidebars = {}
# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/images/favicon.ico"
html_logo = ""
html_last_updated_fmt = "%d %b %Y"

globalnavlinks = {
    "ASDF Projects": "https://www.asdf-format.org",
    "Tutorials": "https://www.asdf-format.org/en/latest/tutorials/index.html",
    "Community": "https://www.asdf-format.org/en/latest/community/index.html",
}

topbanner = ""
for text, link in globalnavlinks.items():
    topbanner += f"<a href={link}>{text}</a>"

html_theme_options = {
    "light_logo": "images/logo-light-mode.png",
    "dark_logo": "images/logo-dark-mode.png",
    "announcement": topbanner,
}

pygments_style = "monokai"
# NB Dark style pygments is furo-specific at this time
pygments_dark_style = "monokai"
# Render inheritance diagrams in SVG
graphviz_output_format = "svg"

graphviz_dot_args = [
    "-Nfontsize=10",
    "-Nfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Efontsize=10",
    "-Efontname=Helvetica Neue, Helvetica, Arial, sans-serif",
    "-Gbgcolor=white",
    "-Gfontsize=10",
    "-Gfontname=Helvetica Neue, Helvetica, Arial, sans-serif",
]

# -- Options for LaTeX output --------------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [("index", project + ".tex", project + " Documentation", author, "manual")]

latex_logo = "_static/images/logo-light-mode.png"


def setup(app):
    app.add_css_file("css/globalnav.css")



import os

import pytest

collect_ignore = ["conf.py"]

@pytest.fixture(autouse=True)
def tmp_cwd(tmp_path):
    """Run doctests in a temporary directory."""
    curdir = os.getcwd()
    try:
        os.chdir(tmp_path)
        yield
    finally:
        os.chdir(curdir)



import numpy as np

import asdf


def test_update_corrupts_stream_data(tmp_path):
    """
    update corrupts stream data
    https://github.com/asdf-format/asdf/issues/1523
    """
    fn = tmp_path / "stream.asdf"

    s = asdf.Stream([3], np.uint8)
    asdf.AsdfFile({"s": s}).write_to(fn)

    with open(fn, "rb+") as f:
        f.seek(0, 2)
        f.write(b"\x01\x02\x03")

    with asdf.open(fn) as af:
        np.testing.assert_array_equal(af["s"], [[1, 2, 3]])

    with asdf.open(fn, mode="rw") as af:
        af["a"] = np.arange(1000)
        af.update()
        np.testing.assert_array_equal(af["s"], [[1, 2, 3]])

    with asdf.open(fn) as af:
        np.testing.assert_array_equal(af["s"], [[1, 2, 3]])



import os

import numpy as np

import asdf


def test_rewrite_file_with_unaccessed_external_blocks_fails(tmp_path):
    """
    Rewriting a file with external blocks fails if arrays are not first accessed

    https://github.com/asdf-format/asdf/issues/1526
    """
    arrs = [np.arange(3) + i for i in range(3)]
    af = asdf.AsdfFile({"arrs": arrs})
    [af.set_array_storage(a, "external") for a in arrs]

    dns = []
    for i in range(2):
        dn = tmp_path / f"d{i}"
        if not os.path.exists(dn):
            os.makedirs(dn)
        dns.append(dn)
    fns = [dn / "test.asdf" for dn in dns]

    # write to d0
    af.write_to(fns[0])

    with asdf.open(fns[0]) as af2:
        af2["arrs"][0] = 42
        # write to d1
        af2.write_to(fns[1])

    assert len(os.listdir(dns[0])) == 4
    assert len(os.listdir(dns[1])) == 3



import asdf


def test_id_in_tree_breaks_ref(tmp_path):
    """
    a dict containing id will break contained References

    https://github.com/asdf-format/asdf/issues/1715
    """
    external_fn = tmp_path / "external.asdf"

    external_tree = {"thing": 42}

    asdf.AsdfFile(external_tree).write_to(external_fn)

    main_fn = tmp_path / "main.asdf"

    af = asdf.AsdfFile({})
    af["id"] = "bogus"
    af["myref"] = {"$ref": "external.asdf#/thing"}
    af.write_to(main_fn)

    with asdf.open(main_fn) as af:
        af.resolve_references()
        assert af["myref"] == 42



import gc

import pytest

from asdf._block.callback import DataCallback
from asdf._block.manager import ReadBlocks


def test_default_attribute():
    class Data:
        def __init__(self, value):
            self.data = value

    blks = ReadBlocks([Data("a"), Data("b")])
    cbs = [DataCallback(0, blks), DataCallback(1, blks)]

    assert cbs[0]() == "a"
    assert cbs[1]() == "b"


def test_attribute_access():
    class Foo:
        def __init__(self, attr, value):
            setattr(self, attr, value)

    blks = ReadBlocks([Foo("a", "foo"), Foo("a", "bar")])
    cb = DataCallback(0, blks)

    assert cb(_attr="a") == "foo"


def test_weakref():
    class Data:
        def __init__(self, value):
            self.data = value

    blks = ReadBlocks([Data("a"), Data("b")])
    cb = DataCallback(0, blks)
    del blks
    gc.collect(2)

    with pytest.raises(OSError, match="Attempt to read block data from missing block"):
        cb()


def test_reassign():
    class Data:
        def __init__(self, value):
            self.data = value

    blks = ReadBlocks([Data("a"), Data("b")])
    cb = DataCallback(0, blks)

    assert cb() == "a"

    blks2 = ReadBlocks([Data("c"), Data("d")])
    cb._reassign(1, blks2)

    assert cb() == "d"



import asdf


def test_info_on_non_tagged_asdf_traverse_object():
    """
    Calling info with a tree containing an object that implements
    __asdf_traverse__ but does not have a _tag results in an
    error

    https://github.com/asdf-format/asdf/issues/1738
    """

    class MyContainer:
        def __init__(self, data):
            self.data = data

        def __asdf_traverse__(self):
            return self.data

    c = MyContainer([1, 2, 3])
    af = asdf.AsdfFile()
    af["c"] = c

    # info should not error out
    af.info()

    # and search should work with the container
    assert af.search(type_=int).paths == ["root['c'][0]", "root['c'][1]", "root['c'][2]"]

    # this should work even if _tag exists (and is not a tag)
    c._tag = {}
    af.info()
    assert af.search(type_=int).paths == ["root['c'][0]", "root['c'][1]", "root['c'][2]"]
    c._tag = "a"
    af.info()
    assert af.search(type_=int).paths == ["root['c'][0]", "root['c'][1]", "root['c'][2]"]



import contextlib
import io
from datetime import datetime

import numpy as np
import pytest
from numpy.testing import assert_array_equal

import asdf
from asdf import config_context, constants, get_config, schema, tagged, util, yamlutil
from asdf.exceptions import AsdfConversionWarning, AsdfWarning, ValidationError
from asdf.extension import TagDefinition
from asdf.testing.helpers import yaml_to_asdf


@contextlib.contextmanager
def tag_reference_extension():
    class TagReference:
        def __init__(self, name, things):
            self.name = name
            self.things = things

    tag_uri = "tag:nowhere.org:custom/tag_reference-1.0.0"
    schema_uri = "http://nowhere.org/schemas/custom/tag_reference-1.0.0"
    tag_def = asdf.extension.TagDefinition(tag_uri, schema_uris=schema_uri)

    class TagReferenceConverter:
        tags = [tag_uri]
        types = [TagReference]

        def to_yaml_tree(self, obj, tag, ctx):
            return {"name": obj.name, "things": obj.things}

        def from_yaml_tree(self, node, tag, ctx):
            return TagReference(node["name"], node["things"])

    class TagReferenceExtension:
        tags = [tag_def]
        extension_uri = "asdf://nowhere.org/extensions/tag_reference-1.0.0"
        converters = [TagReferenceConverter()]

    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: {schema_uri}
title: An example custom type for testing tag references

type: object
properties:
  name:
    type: string
  things:
    $ref: "http://stsci.edu/schemas/asdf/core/ndarray-1.1.0"
required: [name, things]
...
    """

    with config_context() as cfg:
        cfg.add_resource_mapping({schema_uri: tag_schema})
        cfg.add_extension(TagReferenceExtension())
        yield


def test_tagging_scalars():
    class Scalar:
        def __init__(self, value):
            self.value = value

    scalar_tag = "http://somewhere.org/tags/scalar-1.0.0"

    class ScalarConverter:
        tags = [scalar_tag]
        types = [Scalar]

        def to_yaml_tree(self, obj, tag, ctx):
            return obj.value

        def from_yaml_tree(self, node, tag, ctx):
            return Scalar(node)

    class ScalarExtension:
        tags = [scalar_tag]
        converters = [ScalarConverter()]
        extension_uri = "http://somewhere.org/extensions/scalar-1.0.0"

    yaml = f"""
tagged: !<{scalar_tag}>
  m
not_tagged:
  m
    """
    with asdf.config_context() as cfg:
        cfg.add_extension(ScalarExtension())
        buff = yaml_to_asdf(yaml)
        with asdf.open(buff) as ff:
            assert isinstance(ff.tree["tagged"], Scalar)
            assert not isinstance(ff.tree["not_tagged"], Scalar)
            assert isinstance(ff.tree["not_tagged"], str)

            assert ff.tree["tagged"].value == "m"
            assert ff.tree["not_tagged"] == "m"


def test_read_json_schema(test_data_path):
    """Pytest to make sure reading JSON schemas succeeds.

    This was known to fail on Python 3.5 See issue #314 at
    https://github.com/asdf-format/asdf/issues/314 for more details.
    """
    json_schema = test_data_path / "example_schema.json"
    schema_tree = schema.load_schema(json_schema, resolve_references=True)
    schema.check_schema(schema_tree)


def test_load_schema(tmp_path):
    schema_def = """
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "http://stsci.edu/schemas/asdf/nugatory/nugatory-1.0.0"
tag: "tag:stsci.edu:asdf/nugatory/nugatory-1.0.0"

type: object
properties:
  foobar:
      $ref: "../core/ndarray-1.1.0"

required: [foobar]
...
    """
    schema_path = tmp_path / "nugatory.yaml"
    schema_path.write_bytes(schema_def.encode())

    schema_tree = schema.load_schema(str(schema_path), resolve_references=True)
    schema.check_schema(schema_tree)


def test_load_schema_with_file_url(tmp_path):
    schema_def = """
%YAML 1.1
%TAG !asdf! tag:stsci.edu:asdf/
---
$schema: "http://stsci.edu/schemas/asdf/asdf-schema-1.1.0"
id: "http://stsci.edu/schemas/asdf/nugatory/nugatory-1.0.0"
tag: "tag:stsci.edu:asdf/nugatory/nugatory-1.0.0"

type: object
properties:
  foobar:
      $ref: "http://stsci.edu/schemas/asdf/core/ndarray-1.1.0"

required: [foobar]
...
    """
    schema_path = tmp_path / "nugatory.yaml"
    schema_path.write_bytes(schema_def.encode())

    schema_tree = schema.load_schema(str(schema_path), resolve_references=True)
    schema.check_schema(schema_tree)


def test_load_schema_with_asdf_uri_scheme():
    subschema_content = """%YAML 1.1
---
$schema: http://stsci.edu/schemas/asdf/asdf-schema-1.1.0
id: asdf://somewhere.org/schemas/bar

bar:
  type: string
...
"""
    content = """%YAML 1.1
---
$schema: http://stsci.edu/schemas/asdf/asdf-schema-1.1.0
id: asdf://somewhere.org/schemas/foo

definitions:
  local_bar:
    type: string

type: object
properties:
  bar:
    $ref: asdf://somewhere.org/schemas/bar#/bar
  local_bar:
    $ref: '#/definitions/local_bar'
...
"""
    with asdf.config_context() as config:
        config.add_resource_mapping({"asdf://somewhere.org/schemas/foo": content})
        config.add_resource_mapping({"asdf://somewhere.org/schemas/bar": subschema_content})

        schema_tree = schema.load_schema("asdf://somewhere.org/schemas/foo")
        instance = {"bar": "baz", "local_bar": "foz"}
        schema.validate(instance, schema=schema_tree)
        with pytest.raises(ValidationError, match=r".* is not of type .*"):
            schema.validate({"bar": 12}, schema=schema_tree)


def test_load_schema_with_stsci_id():
    """
    This tests the following edge case:
    - schema references a subschema provided by the new extension API
    - subschema URI shares a prefix with one of the old-style extension resolvers
    - resolve_references is enabled

    If we're not careful, the old-style resolver will mangle the URI and
    we won't be able to retrieve the schema content.
    """
    subschema_content = """%YAML 1.1
---
$schema: http://stsci.edu/schemas/asdf/asdf-schema-1.1.0
id: http://stsci.edu/schemas/bar

bar:
  type: string
...
"""
    content = """%YAML 1.1
---
$schema: http://stsci.edu/schemas/asdf/asdf-schema-1.1.0
id: http://stsci.edu/schemas/foo

definitions:
  local_bar:
    type: string

type: object
properties:
  bar:
    $ref: http://stsci.edu/schemas/bar#/bar
  local_bar:
    $ref: '#/definitions/local_bar'
...
"""
    with asdf.config_context() as config:
        config.add_resource_mapping({"http://stsci.edu/schemas/foo": content})
        config.add_resource_mapping({"http://stsci.edu/schemas/bar": subschema_content})

        schema_tree = schema.load_schema("http://stsci.edu/schemas/foo", resolve_references=True)
        instance = {"bar": "baz", "local_bar": "foz"}
        schema.validate(instance, schema=schema_tree)
        with pytest.raises(ValidationError, match=r".* is not of type .*"):
            schema.validate({"bar": 12}, schema=schema_tree)


def test_schema_caching():
    # Make sure that if we request the same URL, we get a different object
    # (despite the caching internal to load_schema).  Changes to a schema
    # dict should not impact other uses of that schema.

    s1 = schema.load_schema("http://stsci.edu/schemas/asdf/core/asdf-1.0.0")
    s2 = schema.load_schema("http://stsci.edu/schemas/asdf/core/asdf-1.0.0")
    assert s1 is not s2


def test_load_schema_from_resource_mapping():
    content = b"""
id: http://somewhere.org/schemas/razmataz-1.0.0
type: object
properties:
  foo:
    type: string
  bar:
    type: boolean
"""

    get_config().add_resource_mapping({"http://somewhere.org/schemas/razmataz-1.0.0": content})

    s = schema.load_schema("http://somewhere.org/schemas/razmataz-1.0.0")

    assert s["id"] == "http://somewhere.org/schemas/razmataz-1.0.0"


def test_flow_style():
    class CustomFlow:
        def __init__(self, a, b):
            self.a = a
            self.b = b

    tag_uri = "http://nowhere.org/tags/custom/custom_flow-1.0.0"

    class CustomFlowConverter:
        tags = [tag_uri]
        types = [CustomFlow]

        def to_yaml_tree(self, obj, tag, ctx):
            return {"a": obj.a, "b": obj.b}

        def from_yaml_tree(self, node, tag, ctx):
            return CustomFlow(node["a"], node["b"])

    schema_uri = "http://nowhere.org/schemas/custom/custom_flow-1.0.0"
    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: "{schema_uri}"
type: object
properties:
  a:
    type: number
  b:
    type: number
flowStyle: block
    """

    tag_def = TagDefinition(tag_uri, schema_uris=[schema_uri])

    class CustomFlowExtension:
        extension_uri = "http://nowhere.org/extensions/custom/custom_flow-1.0.0"
        tags = [tag_def]
        converters = [CustomFlowConverter()]

    with config_context() as cfg:
        cfg.add_extension(CustomFlowExtension())
        cfg.add_resource_mapping({schema_uri: tag_schema})
        buff = io.BytesIO()
        ff = asdf.AsdfFile({"custom_flow": CustomFlow(42, 43)})
        ff.write_to(buff)

        assert b"  a: 42\n  b: 43" in buff.getvalue()


def test_style():
    class CustomStyle:
        def __init__(self, message):
            self.message = message

    tag_uri = "http://nowhere.org/tags/custom/custom_style-1.0.0"

    class CustomStyleConverter:
        tags = [tag_uri]
        types = [CustomStyle]

        def to_yaml_tree(self, obj, tag, ctx):
            return obj.message

        def from_yaml_tree(self, node, tag, ctx):
            return CustomStyle(node)

    schema_uri = "http://nowhere.org/schemas/custom/custom_style-1.0.0"
    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: "{schema_uri}"
type: string
style: literal
    """

    tag_def = TagDefinition(tag_uri, schema_uris=[schema_uri])

    class CustomStyleExtension:
        extension_uri = "http://nowhere.org/extensions/custom/custom_style-1.0.0"
        tags = [tag_def]
        converters = [CustomStyleConverter()]

    with config_context() as cfg:
        cfg.add_extension(CustomStyleExtension())
        cfg.add_resource_mapping({schema_uri: tag_schema})

        tree = {"custom_style": CustomStyle("short")}

        buff = io.BytesIO()
        ff = asdf.AsdfFile(tree)
        ff.write_to(buff)

        assert b"|-\n  short\n" in buff.getvalue()


def test_property_order():
    tree = {"foo": np.ndarray([1, 2, 3])}

    buff = io.BytesIO()
    ff = asdf.AsdfFile(tree)
    ff.write_to(buff)

    ndarray_schema = schema.load_schema("http://stsci.edu/schemas/asdf/core/ndarray-1.1.0")
    property_order = ndarray_schema["anyOf"][1]["propertyOrder"]

    last_index = 0
    for prop in property_order:
        index = buff.getvalue().find(prop.encode("utf-8") + b":")
        if index != -1:
            assert index > last_index
            last_index = index


def test_invalid_nested():
    tag_uri = "http://nowhere.org/tags/custom/custom-1.0.0"
    schema_uri = "http://nowhere.org/schemas/custom/custom-1.0.0"
    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: "{schema_uri}"
type: integer
default: 42
    """

    class Custom:
        def __init__(self, value):
            self.value = value

    class CustomConverter:
        tags = [tag_uri]
        types = [Custom]

        def to_yaml_tree(self, obj, tag, ctx):
            return obj.value

        def from_yaml_tree(self, node, tag, ctx):
            return Custom(node)

    tag_def = TagDefinition(tag_uri, schema_uris=[schema_uri])

    class CustomExtension:
        extension_uri = "http://nowhere.org/extensions/custom/custom-1.0.0"
        tags = [tag_def]
        converters = [CustomConverter()]

    yaml = f"""
custom: !<{tag_uri}>
  foo
    """
    buff = yaml_to_asdf(yaml)
    # This should cause a warning but not an error because without explicitly
    # providing an extension, our custom type will not be recognized and will
    # simply be converted to a raw type.
    with pytest.warns(AsdfConversionWarning, match=tag_uri):
        with asdf.open(buff) as af:
            af["custom"]

    buff.seek(0)
    with config_context() as cfg:
        cfg.add_extension(CustomExtension())
        cfg.add_resource_mapping({schema_uri: tag_schema})
        with (
            pytest.raises(ValidationError, match=r".* is not of type .*"),
            asdf.open(
                buff,
            ),
        ):
            pass

        # Make sure tags get validated inside of other tags that know
        # nothing about them.
        yaml = f"""
array: !core/ndarray-1.1.0
  data: [0, 1, 2]
  custom: !<{tag_uri}>
    foo
        """
        buff = yaml_to_asdf(yaml)
        with (
            pytest.raises(ValidationError, match=r".* is not of type .*"),
            asdf.open(
                buff,
            ),
        ):
            pass


def test_invalid_schema():
    s = {"type": "integer"}
    schema.check_schema(s)

    s = {"type": "foobar"}
    with pytest.raises(ValidationError, match=r".* is not valid under any of the given schemas.*"):
        schema.check_schema(s)


def test_defaults():
    s = {"type": "object", "properties": {"a": {"type": "integer", "default": 42}}}

    t = {}

    cls = schema._create_validator(schema.FILL_DEFAULTS)
    validator = cls(s)
    validator.validate(t)

    assert t["a"] == 42

    cls = schema._create_validator(schema.REMOVE_DEFAULTS)
    validator = cls(s)
    validator.validate(t)

    assert t == {}


def test_default_check_in_schema():
    s = {"type": "object", "properties": {"a": {"type": "integer", "default": "foo"}}}

    with pytest.raises(ValidationError, match=r".* is not of type .*"):
        schema.check_schema(s)

    schema.check_schema(s, validate_default=False)


def test_check_complex_default():
    default_software = tagged.TaggedDict({"name": "asdf", "version": "2.7.0"}, "tag:stsci.edu/asdf/core/software-1.0.0")

    s = {
        "type": "object",
        "properties": {
            "a": {"type": "object", "tag": "tag:stsci.edu/asdf/core/software-1.0.0", "default": default_software},
        },
    }

    schema.check_schema(s)

    s["properties"]["a"]["tag"] = "tag:stsci.edu/asdf/core/ndarray-1.1.0"
    with pytest.raises(ValidationError, match=r"mismatched tags, wanted .*, got .*"):
        schema.check_schema(s)


def test_fill_and_remove_defaults():
    tag_uri = "http://nowhere.org/tags/custom/default-1.0.0"
    schema_uri = "http://nowhere.org/schemas/custom/default-1.0.0"
    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: "{schema_uri}"
type: object
properties:
  a:
    type: integer
    default: 42
  b:
    type: object
    properties:
      c:
        type: integer
        default: 82
  d:
    allOf:
      - type: object
        properties:
          e:
            type: integer
            default: 122
      - type: object
        properties:
          f:
            type: integer
            default: 162
  g:
    anyOf:
      - type: object
        properties:
          h:
            type: integer
            default: 202
      - type: object
        properties:
          i:
            type: integer
            default: 242
  j:
    oneOf:
      - type: object
        properties:
          k:
            type: integer
            default: 282
        required: [k]
        additionalProperties: false
      - type: object
        properties:
          l:
            type: integer
            default: 322
        required: [l]
        additionalProperties: false
    """

    class Default(dict):
        pass

    class DefaultConverter:
        tags = [tag_uri]
        types = [Default]

        def to_yaml_tree(self, obj, tag, ctx):
            return dict(obj)

        def from_yaml_tree(self, node, tag, ctx):
            return Default(**node)

    tag_def = TagDefinition(tag_uri, schema_uris=[schema_uri])

    class DefaultExtension:
        tags = [tag_def]
        converters = [DefaultConverter()]
        extension_uri = "http://nowhere.org/extensions/custom/default-1.0.0"

    with config_context() as cfg:
        # later versions do not fill defaults
        cfg.default_version = "1.5.0"
        cfg.add_extension(DefaultExtension())
        cfg.add_resource_mapping({schema_uri: tag_schema})
        yaml = """
custom: !<http://nowhere.org/tags/custom/default-1.0.0>
  b: {}
  d: {}
  g: {}
  j:
    l: 362
        """
        buff = yaml_to_asdf(yaml, version="1.5.0")
        with asdf.open(buff) as ff:
            assert "a" in ff.tree["custom"]
            assert ff.tree["custom"]["a"] == 42
            assert ff.tree["custom"]["b"]["c"] == 82
            # allOf combiner should fill defaults from all subschemas:
            assert ff.tree["custom"]["d"]["e"] == 122
            assert ff.tree["custom"]["d"]["f"] == 162
            # anyOf combiners should be ignored:
            assert "h" not in ff.tree["custom"]["g"]
            assert "i" not in ff.tree["custom"]["g"]
            # oneOf combiners should be ignored:
            assert "k" not in ff.tree["custom"]["j"]
            assert ff.tree["custom"]["j"]["l"] == 362

        buff.seek(0)
        with config_context() as config:
            config.legacy_fill_schema_defaults = False
            with asdf.open(buff) as ff:
                assert "a" not in ff.tree["custom"]
                assert "c" not in ff.tree["custom"]["b"]
                assert "e" not in ff.tree["custom"]["d"]
                assert "f" not in ff.tree["custom"]["d"]
                assert "h" not in ff.tree["custom"]["g"]
                assert "i" not in ff.tree["custom"]["g"]
                assert "k" not in ff.tree["custom"]["j"]
                assert ff.tree["custom"]["j"]["l"] == 362
                ff.fill_defaults()
                assert "a" in ff.tree["custom"]
                assert ff.tree["custom"]["a"] == 42
                assert "c" in ff.tree["custom"]["b"]
                assert ff.tree["custom"]["b"]["c"] == 82
                assert ff.tree["custom"]["b"]["c"] == 82
                assert ff.tree["custom"]["d"]["e"] == 122
                assert ff.tree["custom"]["d"]["f"] == 162
                assert "h" not in ff.tree["custom"]["g"]
                assert "i" not in ff.tree["custom"]["g"]
                assert "k" not in ff.tree["custom"]["j"]
                assert ff.tree["custom"]["j"]["l"] == 362
                ff.remove_defaults()
                assert "a" not in ff.tree["custom"]
                assert "c" not in ff.tree["custom"]["b"]
                assert "e" not in ff.tree["custom"]["d"]
                assert "f" not in ff.tree["custom"]["d"]
                assert "h" not in ff.tree["custom"]["g"]
                assert "i" not in ff.tree["custom"]["g"]
                assert "k" not in ff.tree["custom"]["j"]
                assert ff.tree["custom"]["j"]["l"] == 362


def test_one_of():
    """
    Covers https://github.com/asdf-format/asdf/issues/809
    """

    class OneOf:
        def __init__(self, value):
            self.value = value

    tag_uri = "http://nowhere.org/custom/one_of-1.0.0"

    class OneOfConverter:
        tags = [tag_uri]
        types = [OneOf]

        def to_yaml_tree(self, obj, tag, ctx):
            return {"value": obj.value}

        def from_yaml_tree(self, node, tag, ctx):
            return OneOf(node["value"])

    schema_uri = "http://nowhere.org/schemas/custom/one_of-1.0.0"
    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: "{schema_uri}"
title: |
  oneOf test schema
oneOf:
  - type: object
    properties:
      value:
        type: number
    required: [value]
    additionalProperties: false

  - type: object
    properties:
      value:
        type: string
    required: [value]
    additionalProperties: false
...
    """

    tag_def = TagDefinition(tag_uri, schema_uris=[schema_uri])

    class OneOfExtension:
        extension_uri = "http://nowhere.org/extensions/custom/one_of-1.0.0"
        tags = [tag_def]
        converters = [OneOfConverter()]

    yaml = f"""
one_of: !<{tag_uri}>
  value: foo
    """

    with config_context() as cfg:
        cfg.add_extension(OneOfExtension())
        cfg.add_resource_mapping({schema_uri: tag_schema})

        buff = yaml_to_asdf(yaml)
        with asdf.open(buff) as ff:
            assert ff["one_of"].value == "foo"


def test_tag_reference_validation():
    yaml = """
custom: !<tag:nowhere.org:custom/tag_reference-1.0.0>
  name:
    "Something"
  things: !core/ndarray-1.1.0
    data: [1, 2, 3]
    """

    with tag_reference_extension():
        buff = yaml_to_asdf(yaml, version="1.6.0")
        with asdf.open(buff) as ff:
            custom = ff.tree["custom"]
            assert custom.name == "Something"
            assert_array_equal(custom.things, [1, 2, 3])


def test_foreign_tag_reference_validation():
    class ForeignTagReference:
        def __init__(self, a):
            self.a = a

    tag_uri = "tag:nowhere.org:custom/foreign_tag_reference-1.0.0"
    schema_uri = "http://nowhere.org/schemas/custom/foreign_tag_reference-1.0.0"
    tag_def = asdf.extension.TagDefinition(tag_uri, schema_uris=schema_uri)

    class ForeignTagReferenceConverter:
        tags = [tag_uri]
        types = [ForeignTagReference]

        def to_yaml_tree(self, obj, tag, ctx):
            return {"a": obj.a}

        def from_yaml_tree(self, node, tag, ctx):
            return ForeignTagReference(node["a"])

    class ForeignTagReferenceExtension:
        tags = [tag_def]
        extension_uri = "asdf://nowhere.org/extensions/foreign_tag_reference-1.0.0"
        converters = [ForeignTagReferenceConverter()]

    tag_schema = f"""
%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: {schema_uri}
title: An example custom type for testing tag references

type: object
properties:
  a:
    # Test foreign tag reference using tag URI
    $ref: "http://nowhere.org/schemas/custom/tag_reference-1.0.0"
required: [a]
...
    """

    yaml = """
custom: !<tag:nowhere.org:custom/foreign_tag_reference-1.0.0>
  a: !<tag:nowhere.org:custom/tag_reference-1.0.0>
    name:
      "Something"
    things: !core/ndarray-1.1.0
      data: [1, 2, 3]
    """

    with tag_reference_extension():
        cfg = asdf.get_config()
        cfg.add_resource_mapping({schema_uri: tag_schema})
        cfg.add_extension(ForeignTagReferenceExtension())

        buff = yaml_to_asdf(yaml, version="1.6.0")
        with asdf.open(buff) as ff:
            a = ff.tree["custom"].a
            assert a.name == "Something"
            assert_array_equal(a.things, [1, 2, 3])


def test_self_reference_resolution(test_data_path):
    s = schema.load_schema(
        test_data_path / "self_referencing-1.0.0.yaml",
        resolve_references=True,
    )
    assert "$ref" not in repr(s)
    assert s["anyOf"][1] == s["anyOf"][0]


def test_schema_resolved_via_entry_points():
    """Test that entry points mappings to core schema works"""
    tag = "tag:stsci.edu:asdf/fits/fits-1.0.0"
    extension_manager = asdf.extension.get_cached_extension_manager(get_config().extensions)
    schema_uris = extension_manager.get_tag_definition(tag).schema_uris
    assert len(schema_uris) > 0
    s = schema.load_schema(schema_uris[0])
    assert s["id"] == schema_uris[0]


@pytest.mark.parametrize("num", [constants.MAX_NUMBER + 1, constants.MIN_NUMBER - 1])
def test_max_min_literals(num):
    msg = r"Integer value .* is too large to safely represent as a literal in ASDF"

    af = asdf.AsdfFile()
    af["test_int"] = num
    with pytest.raises(ValidationError, match=msg):
        af.validate()

    af = asdf.AsdfFile()
    af["test_list"] = [num]
    with pytest.raises(ValidationError, match=msg):
        af.validate()

    af = asdf.AsdfFile()
    af[num] = "test_key"
    with pytest.raises(ValidationError, match=msg):
        af.validate()


@pytest.mark.parametrize("num", [constants.MAX_NUMBER + 1, constants.MIN_NUMBER - 1])
@pytest.mark.parametrize("ttype", ["val", "list", "key"])
def test_max_min_literals_write(num, ttype, tmp_path):
    outfile = tmp_path / "test.asdf"
    af = asdf.AsdfFile()

    # Validation doesn't occur here, so no warning/error will be raised.
    if ttype == "val":
        af.tree["test_int"] = num
    elif ttype == "list":
        af.tree["test_int"] = [num]
    else:
        af.tree[num] = "test_key"

    # Validation will occur on write, though, so detect it.
    msg = r"Integer value .* is too large to safely represent as a literal in ASDF"
    with pytest.raises(ValidationError, match=msg):
        af.write_to(outfile)
    af.close()


@pytest.mark.parametrize("value", [constants.MAX_NUMBER + 1, constants.MIN_NUMBER - 1])
def test_read_large_literal(value):
    yaml = f"integer: {value}"

    buff = yaml_to_asdf(yaml)

    with pytest.warns(AsdfWarning, match=r"Invalid integer literal value"), asdf.open(buff) as af:
        assert af["integer"] == value

    yaml = f"{value}: foo"

    buff = yaml_to_asdf(yaml)

    with pytest.warns(AsdfWarning, match=r"Invalid integer literal value"), asdf.open(buff) as af:
        assert af[value] == "foo"


@pytest.mark.parametrize(
    ("version", "keys"),
    [
        ("1.6.0", ["foo", 42, True]),
        ("1.5.0", ["foo", 42, True, 3.14159, datetime.now(), b"foo", None]),
    ],
)
def test_mapping_supported_key_types(keys, version):
    for key in keys:
        af = asdf.AsdfFile({key: "value"}, version=version)
        buff = io.BytesIO()
        af.write_to(buff)
        buff.seek(0)
        with asdf.open(buff) as af:
            assert af[key] == "value"


@pytest.mark.parametrize(
    ("version", "keys"),
    [
        ("1.6.0", [3.14159, datetime.now(), b"foo", None, ("foo", "bar")]),
    ],
)
def test_mapping_unsupported_key_types(keys, version):
    for key in keys:
        af = asdf.AsdfFile(version=version)
        af[key] = "value"
        with pytest.raises(ValidationError, match=r"Mapping key .* is not permitted"):
            af.validate()


def test_nested_array():
    s = {
        "type": "object",
        "properties": {
            "stuff": {
                "type": "array",
                "items": {
                    "type": "array",
                    "items": [
                        {"type": "integer"},
                        {"type": "string"},
                        {"type": "number"},
                    ],
                    "minItems": 3,
                    "maxItems": 3,
                },
            },
        },
    }

    good = {"stuff": [[1, "hello", 2], [4, "world", 9.7]]}
    schema.validate(good, schema=s)

    bads = [
        {"stuff": [[1, 2, 3]]},
        {"stuff": [12, "dldl"]},
        {"stuff": [[12, "dldl"]]},
        {"stuff": [[1, "hello", 2], [4, 5]]},
        {"stuff": [[1, "hello", 2], [4, 5, 6]]},
    ]

    for b in bads:
        with pytest.raises(ValidationError, match=r"[.* is not of type .*, .* is too short]"):
            schema.validate(b, schema=s)


def test_nested_array_yaml(tmp_path):
    schema_def = """
%YAML 1.1
---
type: object
properties:
  stuff:
    type: array
    items:
      type: array
      items:
        - type: integer
        - type: string
        - type: number
      minItems: 3
      maxItems: 3
...
    """
    schema_path = tmp_path / "nested.yaml"
    schema_path.write_bytes(schema_def.encode())

    schema_tree = schema.load_schema(str(schema_path))
    schema.check_schema(schema_tree)

    good = {"stuff": [[1, "hello", 2], [4, "world", 9.7]]}
    schema.validate(good, schema=schema_tree)

    bads = [
        {"stuff": [[1, 2, 3]]},
        {"stuff": [12, "dldl"]},
        {"stuff": [[12, "dldl"]]},
        {"stuff": [[1, "hello", 2], [4, 5]]},
        {"stuff": [[1, "hello", 2], [4, 5, 6]]},
    ]

    for b in bads:
        with pytest.raises(ValidationError, match=r"[.* is not of type .*, .* is too short]"):
            schema.validate(b, schema=schema_tree)


@pytest.mark.parametrize(
    ("numpy_value", "valid_types"),
    [
        (np.str_("foo"), {"string"}),
        (np.bytes_("foo"), set()),
        (np.float16(3.14), {"number"}),
        (np.float32(3.14159), {"number"}),
        (np.float64(3.14159), {"number"}),
        # Evidently float128 is not available on Windows:
        (getattr(np, "float128", np.float64)(3.14159), {"number"}),
        (np.int8(42), {"number", "integer"}),
        (np.int16(42), {"number", "integer"}),
        (np.int32(42), {"number", "integer"}),
        (np.longlong(42), {"number", "integer"}),
        (np.uint8(42), {"number", "integer"}),
        (np.uint16(42), {"number", "integer"}),
        (np.uint32(42), {"number", "integer"}),
        (np.uint64(42), {"number", "integer"}),
        (np.ulonglong(42), {"number", "integer"}),
    ],
)
def test_numpy_scalar_type_validation(numpy_value, valid_types):
    def _assert_validation(jsonschema_type, expected_valid):
        validator = schema.get_validator(schema={"type": jsonschema_type})
        try:
            validator.validate(numpy_value)
        except ValidationError:
            valid = False
        else:
            valid = True

        if valid is not expected_valid:
            description = "valid" if expected_valid else "invalid"
            msg = (
                f"Expected numpy.{type(numpy_value).__name__} "
                f"to be {description} against jsonschema type "
                f"'{jsonschema_type}'"
            )
            raise AssertionError(msg)

    for jsonschema_type in valid_types:
        _assert_validation(jsonschema_type, True)

    invalid_types = {"string", "number", "integer", "boolean", "null", "object"} - valid_types
    for jsonschema_type in invalid_types:
        _assert_validation(jsonschema_type, False)


def test_validator_visit_repeat_nodes():
    ctx = asdf.AsdfFile()
    node = asdf.tags.core.Software(name="Minesweeper")
    tree = yamlutil.custom_tree_to_tagged_tree({"node": node, "other_node": node, "nested": {"node": node}}, ctx)

    visited_nodes = []

    def _test_validator(validator, value, instance, schema):
        visited_nodes.append(instance)

    validator = schema.get_validator(ctx=ctx, validators=util.HashableDict(type=_test_validator))
    validator.validate(tree)
    assert len(visited_nodes) == 1

    visited_nodes.clear()
    validator = schema.get_validator(validators=util.HashableDict(type=_test_validator), _visit_repeat_nodes=True)
    validator.validate(tree)
    assert len(visited_nodes) == 3


def test_tag_validator():
    content = """%YAML 1.1
---
$schema: http://stsci.edu/schemas/asdf/asdf-schema-1.1.0
id: asdf://somewhere.org/schemas/foo
tag: asdf://somewhere.org/tags/foo
...
"""
    with asdf.config_context() as config:
        config.add_resource_mapping({"asdf://somewhere.org/schemas/foo": content})

        schema_tree = schema.load_schema("asdf://somewhere.org/schemas/foo")
        instance = tagged.TaggedDict(tag="asdf://somewhere.org/tags/foo")
        schema.validate(instance, schema=schema_tree)
        with pytest.raises(ValidationError, match=r"mismatched tags, wanted .*, got .*"):
            schema.validate(tagged.TaggedDict(tag="asdf://somewhere.org/tags/bar"), schema=schema_tree)

    content = """%YAML 1.1
---
$schema: http://stsci.edu/schemas/asdf/asdf-schema-1.1.0
id: asdf://somewhere.org/schemas/bar
tag: asdf://somewhere.org/tags/bar-*
...
"""
    with asdf.config_context() as config:
        config.add_resource_mapping({"asdf://somewhere.org/schemas/bar": content})

        schema_tree = schema.load_schema("asdf://somewhere.org/schemas/bar")
        instance = tagged.TaggedDict(tag="asdf://somewhere.org/tags/bar-2.5")
        schema.validate(instance, schema=schema_tree)
        with pytest.raises(ValidationError, match=r"mismatched tags, wanted .*, got .*"):
            schema.validate(tagged.TaggedDict(tag="asdf://somewhere.org/tags/foo-1.0"), schema=schema_tree)


def test_fail_under_combiner():
    """
    Test that a failed validation under a schema combiner that allows failures
    (oneOf, etc) does not result in an incorrect passing of validation
    which used to be the case due to a bug in how repeat nodes
    were handled.

    See: https://github.com/asdf-format/asdf/issues/2000
    """
    yaml_str = b"""#ASDF 1.0.0
#ASDF_STANDARD 1.6.0
%YAML 1.1
%TAG ! tag:stsci.edu:asdf/
--- !core/asdf-1.1.0
obj: !core/complex-1.0.0 foo
..."""

    schema_id = "http://example.com/failing-1.0.0"
    schema_str = f"""%YAML 1.1
---
$schema: "http://stsci.edu/schemas/yaml-schema/draft-01"
id: {schema_id}
properties:
  obj:
    oneOf:
      - $ref: "http://stsci.edu/schemas/asdf/core/complex-1.0.0"
      - true
..."""

    with asdf.config_context() as cfg:
        cfg.add_resource_mapping({schema_id: schema_str})
        bio = io.BytesIO(yaml_str)
        with pytest.raises(ValidationError, match=r"'foo' does not match"):
            # using lazy_tree here to that the invalid complex is not
            # deserialized (but is still validated)
            with asdf.open(bio, custom_schema=schema_id, lazy_tree=True):
                pass



from itertools import combinations

from asdf.versioning import (
    AsdfVersion,
    default_version,
    supported_versions,
)


def test_default_in_supported_versions():
    assert default_version in supported_versions


def test_version_constructor():
    ver0 = AsdfVersion("1.0.0")
    ver1 = AsdfVersion((1, 0, 0))
    ver2 = AsdfVersion([1, 0, 0])

    assert str(ver0) == "1.0.0"
    assert str(ver1) == "1.0.0"
    assert str(ver2) == "1.0.0"


def test_version_and_version_equality():
    ver0 = AsdfVersion("1.0.0")
    ver1 = AsdfVersion("1.0.0")

    assert ver0 is not ver1
    assert ver0 == ver1
    assert ver1 == ver0
    assert not (ver0 != ver1)
    assert not (ver1 != ver0)


def test_version_and_string_equality():
    version = AsdfVersion("1.0.0")
    string_ver = "1.0.0"

    assert version == string_ver
    assert string_ver == version
    assert not (version != string_ver)
    assert not (string_ver != version)


def test_version_and_tuple_equality():
    version = AsdfVersion("1.0.0")
    tuple_ver = (1, 0, 0)

    assert version == tuple_ver
    assert tuple_ver == version
    assert not (version != tuple_ver)
    assert not (tuple_ver != version)


def test_version_and_version_inequality():
    ver0 = AsdfVersion("1.0.0")
    ver1 = AsdfVersion("1.0.1")
    ver2 = AsdfVersion("1.1.0")
    ver3 = AsdfVersion("1.1.1")
    ver4 = AsdfVersion("2.0.0")
    ver5 = AsdfVersion("2.0.1")
    ver6 = AsdfVersion("2.1.0")
    ver7 = AsdfVersion("2.1.1")

    versions = [ver0, ver1, ver2, ver3, ver4, ver5, ver6, ver7]
    for x, y in combinations(versions, 2):
        assert not (x == y)
        assert x != y

    assert ver0 < ver1 < ver2 < ver3 < ver4 < ver5 < ver6 < ver7
    assert ver7 > ver6 > ver5 > ver4 > ver3 > ver2 > ver1 > ver0
    assert (ver0 < ver1 < ver2 < ver4 < ver3 < ver5 < ver6 < ver7) is False
    assert (ver7 > ver6 > ver5 > ver3 > ver4 > ver2 > ver1 > ver0) is False

    assert ver0 <= ver1 <= ver2 <= ver3 <= ver4 <= ver5 <= ver6 <= ver7
    assert ver7 >= ver6 >= ver5 >= ver4 >= ver3 >= ver2 >= ver1 >= ver0


def test_version_and_string_inequality():
    version = AsdfVersion("2.0.0")

    assert version > "1.0.0"
    assert version > "1.0.1"
    assert version > "1.1.0"
    assert version > "1.1.1"
    assert (version > "2.0.0") is False
    assert (version < "2.0.0") is False
    assert version < "2.0.1"
    assert version < "2.1.0"
    assert version < "2.1.1"

    assert version >= "1.0.0"
    assert version >= "1.0.1"
    assert version >= "1.1.0"
    assert version >= "1.1.1"
    assert version >= "2.0.0"
    assert version <= "2.0.0"
    assert version <= "2.0.1"
    assert version <= "2.1.0"
    assert version <= "2.1.1"

    assert "1.0.0" < version
    assert "1.0.1" < version
    assert "1.1.0" < version
    assert "1.1.1" < version
    assert ("2.0.0" < version) is False
    assert ("2.0.0" > version) is False
    assert "2.0.1" > version
    assert "2.1.0" > version
    assert "2.1.1" > version

    assert "1.0.0" <= version
    assert "1.0.1" <= version
    assert "1.1.0" <= version
    assert "1.1.1" <= version
    assert "2.0.0" <= version
    assert "2.0.0" >= version
    assert "2.0.1" >= version
    assert "2.1.0" >= version
    assert "2.1.1" >= version


def test_version_and_tuple_inequality():
    version = AsdfVersion("2.0.0")

    assert version > (1, 0, 0)
    assert version > (1, 0, 1)
    assert version > (1, 1, 0)
    assert version > (1, 1, 1)
    assert (version > (2, 0, 0)) is False
    assert (version < (2, 0, 0)) is False
    assert version < (2, 0, 1)
    assert version < (2, 1, 0)
    assert version < (2, 1, 1)

    assert version >= (1, 0, 0)
    assert version >= (1, 0, 1)
    assert version >= (1, 1, 0)
    assert version >= (1, 1, 1)
    assert version >= (2, 0, 0)
    assert version <= (2, 0, 0)
    assert version <= (2, 0, 1)
    assert version <= (2, 1, 0)
    assert version <= (2, 1, 1)

    assert (1, 0, 0) < version
    assert (1, 0, 1) < version
    assert (1, 1, 0) < version
    assert (1, 1, 1) < version
    assert ((2, 0, 0) < version) is False
    assert ((2, 0, 0) > version) is False
    assert (2, 0, 1) > version
    assert (2, 1, 0) > version
    assert (2, 1, 1) > version

    assert (1, 0, 0) <= version
    assert (1, 0, 1) <= version
    assert (1, 1, 0) <= version
    assert (1, 1, 1) <= version
    assert (2, 0, 0) <= version
    assert (2, 0, 0) >= version
    assert (2, 0, 1) >= version
    assert (2, 1, 0) >= version
    assert (2, 1, 1) >= version



import numpy as np

import asdf


def test_update_with_memmapped_data_can_make_view_data_invalid(tmp_path):
    """
    Calling update with memmapped data can create invalid data in memmap views

    https://github.com/asdf-format/asdf/issues/1530

    A view of a memmapped array can return invalid data or segfault
    after an update
    """
    fn = tmp_path / "test.asdf"
    a = np.zeros(10, dtype="uint8")
    b = np.ones(10, dtype="uint8")
    ov = a[:3]

    af = asdf.AsdfFile({"a": a, "b": b})
    af.write_to(fn)

    with asdf.open(fn, mode="rw", memmap=True) as af:
        va = af["a"][:3]
        np.testing.assert_array_equal(a, af["a"])
        np.testing.assert_array_equal(b, af["b"])
        np.testing.assert_array_equal(va, ov)
        af["c"] = "a" * 10000
        af.update()
        np.testing.assert_array_equal(a, af["a"])
        np.testing.assert_array_equal(b, af["b"])
        # the view of 'a' taken above ('va') keeps the original memmap open
        # and is not a valid view of af['a'] (as this now differs from the
        # af['a'] used to generate the view).
        assert not np.all(va == ov)



import numpy as np

from asdf import constants

from . import io as bio


class WriteBlock:
    """
    Data and compression options needed to write an ASDF block.
    """

    def __init__(self, data, compression=None, compression_kwargs=None):
        self._data = data
        self.compression = compression
        self.compression_kwargs = compression_kwargs

    @property
    def data(self):
        if callable(self._data):
            return self._data()
        return self._data

    @property
    def data_bytes(self):
        data = self.data
        if data is not None:
            return np.ndarray(-1, np.uint8, data.ravel(order="K").data)
        return np.ndarray(0, np.uint8)


def write_blocks(fd, blocks, padding=False, streamed_block=None, write_index=True, write_checksums=True):
    """
    Write a list of WriteBlocks to a file

    Parameters
    ----------
    fd : file or generic_io.GenericIO
        File to write to. Writing will start at the current position.

    blocks : list of WriteBlock
        List of WriteBlock instances used to get the data and options
        to write to each ASDF block.

    padding : bool or float, optional, default False
        If False, add no padding bytes between blocks. If True
        add some default amount of padding. If a float, add
        a number of padding bytes based off a ratio of the data
        size.
        See ``asdf._block.io.write_block`` ``padding`` argument for
        more details.

    streamed_block : WriteBlock, optional
        If provided (not None) include this WriteBlock as
        the final block in the file and mark it as a streamed
        block.

    write_index : bool, optional, default True
        If True, include a block index at the end of the file.
        If a streamed_block is provided (or the file is not
        seekable) no block index will be written.

    write_checksums: bool, optional
        Compute and write block checksums to the file.

    Returns
    -------
    offsets : list of int
        Byte offsets (from the start of the file) where each
        block was written (this is the start of the block magic
        bytes for each block). This list includes the offset of
        the streamed_block if it was provided.
        If the file written to is not seekable these offsets
        will all be None.

    headers : list of dict
        Headers written for each block (including the streamed_block
        if it was provided).
    """
    # some non-seekable files return a valid `tell` result
    # others can raise an exception, others might always
    # return 0. See relevant issues:
    # https://github.com/asdf-format/asdf/issues/1545
    # https://github.com/asdf-format/asdf/issues/1552
    # https://github.com/asdf-format/asdf/issues/1542
    # to enable writing a block index for all valid files
    # we will wrap tell to return None on an error

    def tell():
        try:
            return fd.tell()
        except OSError:
            return None

    offsets = []
    headers = []
    for blk in blocks:
        offsets.append(tell())
        fd.write(constants.BLOCK_MAGIC)
        headers.append(
            bio.write_block(
                fd,
                blk.data_bytes,
                compression_kwargs=blk.compression_kwargs,
                padding=padding,
                compression=blk.compression,
                write_checksum=write_checksums,
            )
        )
    if streamed_block is not None:
        offsets.append(tell())
        fd.write(constants.BLOCK_MAGIC)
        headers.append(bio.write_block(fd, streamed_block.data_bytes, stream=True, write_checksum=write_checksums))

    # os.pipe on windows returns a file-like object
    # that reports as seekable but tell always returns 0
    # https://github.com/asdf-format/asdf/issues/1545
    # when all offsets are 0 replace them with all Nones
    if all(o == 0 for o in offsets):
        offsets = [None for _ in offsets]

    # only write a block index if all conditions are met
    if streamed_block is None and write_index and len(offsets) and all(o is not None for o in offsets):
        bio.write_block_index(fd, offsets)
    return offsets, headers



"""
Defragment command.
"""

import warnings

import asdf
from asdf import AsdfFile

from .main import Command

__all__ = ["defragment"]


class Defragment(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "defragment",
            help="Defragment an ASDF file..",
            description="""Removes any unused blocks and unused space.""",
        )

        parser.add_argument("filename", nargs=1, help="""The ASDF file to collect.""")
        parser.add_argument("--output", "-o", type=str, nargs="?", help="""The name of the output file.""")
        parser.add_argument(
            "--resolve-references",
            "-r",
            action="store_true",
            help="""Resolve all references and store them directly in
            the output file.""",
        )
        parser.add_argument(
            "--compress",
            "-c",
            type=str,
            nargs="?",
            choices=["zlib", "bzp2", "lz4"],
            help="""Compress blocks using one of "zlib", "bzp2" or "lz4".""",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        return defragment(args.filename[0], args.output, args.resolve_references, args.compress)


def defragment(input_, output=None, resolve_references=False, compress=None):
    """
    Defragment a given ASDF file.

    Parameters
    ----------
    input_ : str or file-like object
        The input file.

    output : str of file-like object
        The output file.

    resolve_references : bool, optional
        If `True` resolve all external references before saving.

    compress : str, optional
        Compression to use.
    """
    warnings.warn(
        "defragment is deprecated. Use implode instead",
        UserWarning,
    )
    with asdf.open(input_) as ff:
        ff2 = AsdfFile(ff)
        if resolve_references:
            ff2.resolve_references()
        ff2.write_to(output, all_array_storage="internal", all_array_compression=compress)



from asdf.extension import ManifestExtension
from asdf.versioning import get_supported_core_schema_versions

from ._converters.complex import ComplexConverter
from ._converters.constant import ConstantConverter
from ._converters.external_reference import ExternalArrayReferenceConverter
from ._converters.integer import IntegerConverter
from ._converters.ndarray import NDArrayConverter
from ._converters.reference import ReferenceConverter
from ._converters.tree import (
    AsdfObjectConverter,
    ExtensionMetadataConverter,
    HistoryEntryConverter,
    SoftwareConverter,
    SubclassMetadataConverter,
)
from ._validators import ndarray

CONVERTERS = [
    ComplexConverter(),
    ConstantConverter(),
    ExternalArrayReferenceConverter(),
    AsdfObjectConverter(),
    ExtensionMetadataConverter(),
    HistoryEntryConverter(),
    IntegerConverter(),
    SoftwareConverter(),
    SubclassMetadataConverter(),
    ReferenceConverter(),
    NDArrayConverter(),
]


VALIDATORS = [
    ndarray.NdimValidator(),
    ndarray.MaxNdimValidator(),
    ndarray.DatatypeValidator(),
]


MANIFEST_URIS = [
    f"asdf://asdf-format.org/core/manifests/core-{version}" for version in get_supported_core_schema_versions()
]

EXTENSIONS = [
    ManifestExtension.from_uri(
        u, converters=CONVERTERS, validators=VALIDATORS, legacy_class_names=["asdf.extension.BuiltinExtension"]
    )
    for u in MANIFEST_URIS
]



import abc


class Validator(abc.ABC):
    """
    Abstract base class for plugins that handle custom validators
    in ASDF schemas.
    """

    @abc.abstractproperty
    def schema_property(self):
        """
        Name of the schema property used to invoke this validator.
        """

    @abc.abstractproperty
    def tags(self):
        """
        Get the YAML tags that are appropriate to this validator.
        URI patterns are permitted, see `asdf.util.uri_match` for details.

        Returns
        -------
        iterable of str
            Tag URIs or URI patterns.
        """

    @abc.abstractmethod
    def validate(self, schema_property_value, node, schema):
        """
        Validate the given node from the ASDF tree.

        Parameters
        ----------
        schema_property_value : object
            The value assigned to the schema property associated with this
            valdiator.

        node : asdf.tagged.Tagged
            A tagged node from the tree.  Guaranteed to bear a tag that
            matches one of the URIs returned by this validator's tags property.

        schema : dict
            The schema object that contains the property that triggered
            the validation.  Typically implementations of this method do
            not need to make use of this object, but sometimes the behavior
            of a validator depends on other schema properties.  An example is
            the built-in "additionalProperties" property, which needs to know
            the contents of the "properties" property in order to determine
            which node properties are additional.

        Yields
        ------
        asdf.exceptions.ValidationError
            Yield an instance of ValidationError for each error present in the node.
        """



"""
Contains commands for dealing with exploded and imploded forms.
"""

import os

import asdf
from asdf import AsdfFile

from .main import Command

__all__ = ["to_yaml"]


class ToYaml(Command):
    @classmethod
    def setup_arguments(cls, subparsers):
        parser = subparsers.add_parser(
            "to_yaml",
            help="Convert as ASDF file to pure YAML.",
            description="""Convert all data to inline YAML so the ASDF
            file contains no binary blocks.""",
        )

        parser.add_argument("filename", nargs=1, help="""The ASDF file to convert to YAML.""")
        parser.add_argument(
            "--output",
            "-o",
            type=str,
            nargs="?",
            help="""The name of the output file.  If not provided, it
            will be the name of the input file with a '.yaml' extension.""",
        )
        parser.add_argument(
            "--resolve-references",
            "-r",
            action="store_true",
            help="""Resolve all references and store them directly in
            the output file.""",
        )

        parser.set_defaults(func=cls.run)

        return parser

    @classmethod
    def run(cls, args):
        return to_yaml(args.filename[0], args.output, args.resolve_references)


def to_yaml(input_, output=None, resolve_references=False):
    """
    Implode a given ASDF file, which may reference external data, back
    into a single ASDF file.

    Parameters
    ----------
    input_ : str or file-like object
        The input file.

    output : str of file-like object
        The output file.

    resolve_references : bool, optional
        If `True` resolve all external references before saving.
    """
    if output is None:
        base, _ = os.path.splitext(input_)
        output = base + ".yaml"
    with asdf.open(input_) as ff:
        ff2 = AsdfFile(ff)
        if resolve_references:
            ff2.resolve_references()
        ff2.write_to(output, all_array_storage="inline")



"""
Low-level input/output routines for AsdfFile instances
"""

import contextlib
import io
import pathlib

from . import _version, constants, generic_io, versioning, yamlutil
from ._block.manager import Manager as BlockManager
from .exceptions import DelimiterNotFoundError
from .tags.core import Software


def get_asdf_library_info():
    """
    Get information about asdf to include in the asdf_library entry
    in the Tree.
    """
    return Software(
        {
            "name": "asdf",
            "version": _version.version,
            "homepage": "http://github.com/asdf-format/asdf",
            "author": "The ASDF Developers",
        },
    )


def parse_header_line(line):
    """
    Parses the header line in a ASDF file to obtain the ASDF version.
    """
    parts = line.split()
    if len(parts) != 2 or parts[0] != constants.ASDF_MAGIC:
        msg = "Does not appear to be a ASDF file."
        raise ValueError(msg)

    try:
        version = versioning.AsdfVersion(parts[1].decode("ascii"))
    except ValueError as err:
        msg = f"Unparsable version in ASDF file: {parts[1]}"
        raise ValueError(msg) from err

    if version != versioning._FILE_FORMAT_VERSION:
        msg = f"Unsupported ASDF file format version {version}"
        raise ValueError(msg)

    return version


def read_header_line(gf):
    try:
        header_line = gf.read_until(b"\r?\n", 2, "newline", include=True)
    except DelimiterNotFoundError as e:
        msg = "Does not appear to be a ASDF file."
        raise ValueError(msg) from e

    return parse_header_line(header_line)


def read_comment_section(fd):
    """
    Reads the comment section, between the header line and the
    Tree or first block.
    """
    content = fd.read_until(
        b"(%YAML)|(" + constants.BLOCK_MAGIC + b")",
        5,
        "start of content",
        include=False,
        exception=False,
    )

    comments = []

    lines = content.splitlines()
    for line in lines:
        if not line.startswith(b"#"):
            msg = "Invalid content between header and tree"
            raise ValueError(msg)
        comments.append(line[1:].strip())

    return comments


def find_asdf_version_in_comments(comments, default=None):
    for comment in comments:
        parts = comment.split()
        if len(parts) == 2 and parts[0] == constants.ASDF_STANDARD_COMMENT:
            try:
                version = versioning.AsdfVersion(parts[1].decode("ascii"))
            except ValueError:
                pass
            else:
                return version

    return default


@contextlib.contextmanager
def maybe_close(fd, mode=None, uri=None):
    if mode not in [None, "r", "rw"]:
        msg = f"Unrecognized asdf mode '{mode}'. Must be either 'r' or 'rw'"
        raise ValueError(msg)

    if isinstance(fd, (str, pathlib.Path)):
        mode = mode or "r"
        generic_file = generic_io.get_file(fd, mode=mode, uri=uri)
        try:
            yield generic_file
        except Exception as e:
            generic_file.close()
            raise e
    elif isinstance(fd, generic_io.GenericFile):
        yield fd
    else:
        if mode is None:
            # infer from fd
            if isinstance(fd, io.IOBase):
                mode = "rw" if fd.writable() else "r"
            else:
                mode = "r"
        yield generic_io.get_file(fd, mode=mode, uri=uri)


def read_tree_and_blocks(gf, lazy_load, memmap, validate_checksums):
    token = gf.read(4)
    tree = None
    blocks = BlockManager(uri=gf.uri, lazy_load=lazy_load, memmap=memmap, validate_checksums=validate_checksums)
    if token == b"%YAM":
        reader = gf.reader_until(
            constants.YAML_END_MARKER_REGEX,
            7,
            "End of YAML marker",
            include=True,
            initial_content=token,
        )
        tree = yamlutil.load_tree(reader)
        blocks.read(gf, after_magic=False)
    elif token == constants.BLOCK_MAGIC:
        blocks.read(gf, after_magic=True)
    elif token != b"":
        msg = "ASDF file appears to contain garbage after header."
        raise OSError(msg)

    return tree, blocks


def open_asdf(fd, uri=None, mode=None, lazy_load=True, memmap=False, validate_checksums=False):
    with maybe_close(fd, mode, uri) as generic_file:
        file_format_version = read_header_line(generic_file)
        comments = read_comment_section(generic_file)
        tree, blocks = read_tree_and_blocks(generic_file, lazy_load, memmap, validate_checksums)
        return file_format_version, comments, tree, blocks



"""
Submodule for reading and writing ASDF blocks.

The primary interface to this submodule is ``_block.manager.Manager``
that in some ways mimics the older ``BlockManager``. An instance
of ``Manager`` will be created by each `asdf.AsdfFile` instance.

Internally, this submodule is broken up into:
    - low-level:
        - ``io``: functions for reading and writing blocks
        - ``key``: ``Key`` used to implement ``Store`` (see below)
        - ``store``: ``Store`` special key-value store for indexing blocks
    - medium-level:
        - ``reader``:  ``ReadBlock`` and ``read_blocks``
        - ``writer``:  ``WriteBlock`` and ``write_blocks``
        - ``callback``: ``DataCallback`` for reading block data
        - ``external``: ``ExternalBlockCache`` for reading external blocks
        - ``options``: ``Options`` controlling block storage
    - high-level:
        - ``manager``: ``Manager`` and associated classes


The low-level ``io`` functions are responsible for reading and writing
bytes compatible with the block format defined in the ASDF specification.
These should be compatible with as wide a variety of file formats as possible
including files that are:
    - seekable and non-seekable
    - memory mappable
    - accessed from a remote server
    - stored in memory
    - etc

To help organize ASDF block data the ``key`` and ``store`` submodules
provide a special key-value store, ``Store``. ``Store`` uses ``Key``
instances to tie the lifetime of values to the lifetime of objects
in the ASDF tree (without keeping references to the objects) and
allows non-hashable objects to be used as keys. See the ``key``
submodule docstring for more details. One usage of ``Store`` is
for managing ASDF block ``Options``. ``Options`` determine where
and how array data will be written and a single ``Options`` instance
might be associated with several arrays within the ASDF tree
(if the arrays share the same base array). By using a ``Key`` generated
with the base array the block ``Options`` can be stored in a ``Store``
without keeping a reference to the base array and these ``Options``
will be made unavailable if the base array is garbage collected (so
they are not inapproriately assigned to a new array).

The medium-level submodules ``reader`` and ``writer`` each define
a helper class and function for reading or writing blocks:
    - ``ReadBlock`` and ``WriteBlock``
    - ``read_blocks`` and ``write_blocks``
These abstract some of the complexity of reading and writing blocks
using the low-level API and are the primary means by which the ``Manager``
reads and writes ASDF blocks. Reading of external blocks by the ``Manager``
requires some special handling which is contained in the ``external``
submodule.

To allow for lazy-loading of ASDF block data, ``callback`` defines
``DataCallback`` which allows reading block data even after the blocks
have been rearranged following an update-in-place.
"""



from asdf.tags.core.external_reference import ExternalArrayReference
from asdf.testing.helpers import roundtrip_object


def test_roundtrip_external_array(tmp_path):
    ref = ExternalArrayReference("./nonexistent.fits", 1, "np.float64", (100, 100))

    result = roundtrip_object(ref)

    assert result == ref



"""
Compressor is an interface to implement extensions to the compression
module. Extensions will typically subclass the Compressor ABC and
provide that subclass as a setuptools entry point.

Note that this interface has similar patterns to Converter.  This
interface is designed for compression and decompression of ASDF
binary array blocks, while Converter is designed for serialization
of custom Python types into the YAML tree.
"""

import abc


class Compressor(abc.ABC):
    """
    Abstract base class for plugins that compress binary data.

    Implementing classes must provide the ``labels`` property, and
    at least one of the `compress()` and `decompress()` methods.
    May also provide a constructor.
    """

    @classmethod
    def __subclasshook__(cls, class_):
        if cls is Compressor:
            return hasattr(class_, "label") and (hasattr(class_, "compress") or hasattr(class_, "decompress"))
        return NotImplemented  # pragma: no cover

    @property
    @abc.abstractmethod
    def label(self):
        """
        Get the 4-byte label identifying this compression

        Returns
        -------
        label : bytes
            The compression label
        """

    def compress(self, data, **kwargs):
        """
        Compress ``data``, yielding the results. The yield may be
        block-by-block, or all at once.

        Parameters
        ----------
        data : memoryview
            The data to compress. Must be contiguous and 1D, with
            the underlying ``itemsize`` preserved.
        **kwargs
            Keyword arguments to be passed to the underlying compression
            function

        Yields
        ------
        compressed : bytes-like
            A block of compressed data
        """
        raise NotImplementedError

    def decompress(self, data, out, **kwargs):
        """
        Decompress ``data``, writing the result into ``out``.

        Parameters
        ----------
        data : Iterable of bytes-like
            An Iterable of bytes-like objects containing chunks
            of compressed data.
        out : read-write bytes-like
            A contiguous, 1D output array, of equal or greater length
            than the decompressed data.
        **kwargs
            Keyword arguments to be passed to the underlying decompression
            function

        Returns
        -------
        nbytes : int
            The number of bytes written to ``out``
        """
        raise NotImplementedError



import numpy as np

import asdf


def test_update_fails_after_write_to(tmp_path):
    """
    Calling update after write_to fails

    https://github.com/asdf-format/asdf/issues/1505
    """
    fn1 = tmp_path / "test1.asdf"
    fn2 = tmp_path / "test2.asdf"

    tree = {"a": np.zeros(3), "b": np.ones(3)}
    af = asdf.AsdfFile(tree)

    af.write_to(fn1)

    with asdf.open(fn1, mode="rw") as af:
        af["a"] = None
        af.write_to(fn2)
        af.update()



import pytest

import asdf


@pytest.fixture(scope="module")
def ctx_asdf_file():
    return asdf.AsdfFile()


@pytest.fixture()
def tagged_tree(tree, ctx_asdf_file):
    return asdf.yamlutil.custom_tree_to_tagged_tree(tree, ctx_asdf_file)


def test_custom_tree_to_tagged_tree(tree, ctx_asdf_file, benchmark):
    benchmark(asdf.yamlutil.custom_tree_to_tagged_tree, tree, ctx_asdf_file)


def test_tagged_tree_to_tagged_tree(tagged_tree, ctx_asdf_file, benchmark):
    benchmark(asdf.yamlutil.tagged_tree_to_custom_tree, tagged_tree, ctx_asdf_file)



"""
A hashable Key class that provides a means for tracking
the lifetime of objects to associate objects with
blocks, options and other parts of an asdf file.

This Key is meant to replace uses of id(obj) which in
previous code was used to store settings (like block
array storage). The use of id was problematic as
an object might be deallocated (if it is removed from
the tree and all other references freed) and a new
object of the same type might occupy the same location
in memory and result in the same id. This could result
in options originally associated with the first object
being incorrectly assigned to the new object.

At it's core, Key, uses a weak reference to the object
which can be checked to see if the object is still
in memory.

Instances of this class will be provided to extension
code (see ``SerializationContext.generate_block_key``)
as Converters will need to resupply these keys
on rewrites (updates) to allow asdf to reassociate
objects and blocks. To discourage modifications
of these Key instances all methods and attributes
are private.
"""

import weakref


class Key:
    _next = 0

    @classmethod
    def _next_key(cls):
        key = cls._next
        cls._next += 1
        return key

    def __init__(self, obj=None, _key=None):
        if _key is None:
            _key = Key._next_key()
        self._key = _key
        self._ref = None
        if obj is not None:
            self._assign_object(obj)

    def _is_valid(self):
        if self._ref is None:
            return False
        r = self._ref()
        if r is None:
            return False
        return True

    def __hash__(self):
        return self._key

    def _assign_object(self, obj):
        self._ref = weakref.ref(obj)

    def _matches_object(self, obj):
        if self._ref is None:
            return False
        r = self._ref()
        if r is None:
            return False
        return r is obj

    def __eq__(self, other):
        if not isinstance(other, Key):
            return NotImplemented
        if self._key != other._key:
            return False
        if not self._is_valid():
            return False
        return other._matches_object(self._ref())

    def __copy__(self):
        obj = self._ref if self._ref is None else self._ref()
        return type(self)(obj, self._key)



import io

import pytest

import asdf


@pytest.mark.xfail(reason="Fix will require more major changes to generic_io")
def test_invalid_seek_and_read_from_closed_memoryio():
    """
    Seek and read from closed MemoryIO

    https://github.com/asdf-format/asdf/issues/1539
    """
    b = io.BytesIO()
    b.write(b"\0" * 10)
    b.seek(0)
    f = asdf.generic_io.get_file(b)
    f.close()
    with pytest.raises(IOError, match=r"I/O operation on closed file\."):
        f.read_into_array(10)
    assert b.tell() == 0



import bz2
import struct
import warnings
import zlib

import numpy as np

from .config import get_config
from .exceptions import AsdfWarning


def validate(compression):
    """
    Validate the compression string.

    Parameters
    ----------
    compression : str, bytes or None

    Returns
    -------
    compression : str or None
        In canonical form.

    Raises
    ------
    ValueError
    """
    if not compression or compression == b"\0\0\0\0":
        return None

    if isinstance(compression, bytes):
        compression = compression.decode("ascii")

    compression = compression.strip("\0")

    builtin_labels = ["zlib", "bzp2", "lz4", "input"]
    ext_labels = _get_all_compression_extension_labels()
    all_labels = ext_labels + builtin_labels

    # An extension is allowed to override a builtin compression or another extension,
    # but let's warn the user of this.
    # TODO: is this the desired behavior?
    for i, label in enumerate(all_labels):
        if label in all_labels[i + 1 :]:
            warnings.warn(f'Found more than one compressor for "{label}"', AsdfWarning)

    if compression not in all_labels:
        msg = f"Supported compression types are: {all_labels}, not '{compression}'"
        raise ValueError(msg)

    return compression


class Lz4Compressor:
    def __init__(self):
        try:
            import lz4.block
        except ImportError as err:
            msg = (
                "lz4 library in not installed in your Python environment, "
                "therefore the compressed block in this ASDF file "
                "can not be decompressed."
            )
            raise ImportError(msg) from err

        self._api = lz4.block

    def compress(self, data, **kwargs):
        kwargs["mode"] = kwargs.get("mode", "default")
        compression_block_size = kwargs.pop("compression_block_size", 1 << 22)

        nelem = compression_block_size // data.itemsize
        for i in range(0, len(data), nelem):
            _output = self._api.compress(data[i : i + nelem], **kwargs)
            header = struct.pack("!I", len(_output))
            yield header + _output

    def decompress(self, blocks, out, **kwargs):
        _size = 0
        _pos = 0
        _partial_len = b""
        _buffer = None
        bytesout = 0

        for block in blocks:
            cast = "c"
            blk = memoryview(block).cast(cast)  # don't copy on slice

            while len(blk):
                if not _size:
                    # Don't know the (compressed) length of this block yet
                    if len(_partial_len) + len(blk) < 4:
                        _partial_len += blk
                        break  # we've exhausted the block
                    if _partial_len:
                        # If we started to fill a len key, finish filling it
                        remaining = 4 - len(_partial_len)
                        if remaining:
                            _partial_len += blk[:remaining]
                            blk = blk[remaining:]
                        _size = struct.unpack("!I", _partial_len)[0]
                        _partial_len = b""
                    else:
                        # Otherwise just read the len key directly
                        _size = struct.unpack("!I", blk[:4])[0]
                        blk = blk[4:]

                if len(blk) < _size or _buffer is not None:
                    # If we have a partial block, or we're already filling a buffer, use the buffer
                    if _buffer is None:
                        _buffer = np.empty(
                            _size,
                            dtype=np.byte,
                        )  # use numpy instead of bytearray so we can avoid zero initialization
                        _pos = 0
                    newbytes = min(_size - _pos, len(blk))  # don't fill past the buffer len!
                    _buffer[_pos : _pos + newbytes] = np.frombuffer(blk[:newbytes], dtype=np.byte)
                    _pos += newbytes
                    blk = blk[newbytes:]

                    if _pos == _size:
                        _out = self._api.decompress(_buffer, return_bytearray=True, **kwargs)
                        out[bytesout : bytesout + len(_out)] = _out
                        bytesout += len(_out)
                        _buffer = None
                        _size = 0
                else:
                    # We have at least one full block
                    _out = self._api.decompress(memoryview(blk[:_size]), return_bytearray=True, **kwargs)
                    out[bytesout : bytesout + len(_out)] = _out
                    bytesout += len(_out)
                    blk = blk[_size:]
                    _size = 0

        return bytesout


class ZlibCompressor:
    def compress(self, data, **kwargs):
        comp = zlib.compress(data, **kwargs)
        yield comp

    def decompress(self, blocks, out, **kwargs):
        decompressor = zlib.decompressobj(**kwargs)

        i = 0
        for block in blocks:
            decomp = decompressor.decompress(block)
            out[i : i + len(decomp)] = decomp
            i += len(decomp)
        return i


class Bzp2Compressor:
    def compress(self, data, **kwargs):
        comp = bz2.compress(data, **kwargs)
        yield comp

    def decompress(self, blocks, out, **kwargs):
        decompressor = bz2.BZ2Decompressor(**kwargs)

        i = 0
        for block in blocks:
            decomp = decompressor.decompress(block)
            out[i : i + len(decomp)] = decomp
            i += len(decomp)
        return i


def _get_compressor_from_extensions(compression, return_extension=False):
    """
    Look at the loaded ASDF extensions and return the first one (if any)
    that can handle this type of compression.
    `return_extension` can be used to return corresponding extension for bookkeeping purposes.
    Returns None if no match found.
    """
    # TODO: in ASDF 3, this will be done by the ExtensionManager
    extensions = get_config().extensions

    for ext in extensions:
        for comp in ext.compressors:
            if compression == comp.label.decode("ascii"):
                if return_extension:
                    return comp, ext

                return comp

    return None


def _get_all_compression_extension_labels():
    """
    Get the list of compression labels supported via extensions
    """
    # TODO: in ASDF 3, this will be done by the ExtensionManager
    labels = []
    extensions = get_config().extensions
    for ext in extensions:
        for comp in ext.compressors:
            labels += [comp.label.decode("ascii")]
    return labels


def _get_compressor(label):
    ext_comp = _get_compressor_from_extensions(label)

    if ext_comp is not None:
        # Use an extension before builtins
        comp = ext_comp
    elif label == "zlib":
        comp = ZlibCompressor()
    elif label == "bzp2":
        comp = Bzp2Compressor()
    elif label == "lz4":
        comp = Lz4Compressor()
    else:
        msg = f"Unknown compression type: '{label}'"
        raise ValueError(msg)

    return comp


def to_compression_header(compression):
    """
    Converts a compression string to the four byte field in a block
    header.
    """
    if not compression:
        return b"\0\0\0\0"

    if isinstance(compression, str):
        return compression.encode("ascii")

    return compression


def decompress(fd, used_size, data_size, compression, config=None):
    """
    Decompress binary data in a file

    Parameters
    ----------
    fd : generic_io.GenericIO object
        The file to read the compressed data from.

    used_size : int
        The size of the compressed data

    data_size : int
        The size of the uncompressed data

    compression : str
        The compression type used.

    config : dict or None, optional
        Any kwarg parameters to pass to the underlying decompression
        function

    Returns
    -------
    array : numpy.array
         A flat uint8 containing the decompressed data.
    """
    buffer = np.empty((data_size,), np.uint8)

    compression = validate(compression)
    decoder = _get_compressor(compression)
    if config is None:
        config = {}

    blocks = fd.read_blocks(used_size)  # data is a generator
    len_decoded = decoder.decompress(blocks, out=buffer.data, **config)

    if len_decoded != data_size:
        msg = "Decompressed data wrong size"
        raise ValueError(msg)

    return buffer


def compress(fd, data, compression, config=None):
    """
    Compress array data and write to a file.

    Parameters
    ----------
    fd : generic_io.GenericIO object
        The file to write to.

    data : buffer
        The buffer of uncompressed data.

    compression : str
        The type of compression to use.

    config : dict or None, optional
        Any kwarg parameters to pass to the underlying compression
        function
    """
    compression = validate(compression)
    encoder = _get_compressor(compression)
    if config is None:
        config = {}

    # Get a contiguous, 1D memoryview of the underlying data, preserving data.itemsize
    # - contiguous: because we may not want to assume that all compressors can handle arbitrary strides
    # - 1D: so that len(data) works, not just data.nbytes
    # - itemsize: should preserve data.itemsize for compressors that want to use the record size
    # - memoryview: don't incur the expense of a memcpy, such as with tobytes()
    data = memoryview(data)
    if not data.contiguous:
        data = memoryview(data.tobytes())  # make a contiguous copy
    data = memoryview(np.frombuffer(data, dtype=data.format))  # get a 1D array that preserves byteorder
    if not data.contiguous:
        # the data will be contiguous by construction, but better safe than sorry!
        raise ValueError(data.contiguous)

    compressed = encoder.compress(data, **config)
    # Write block by block
    for comp in compressed:
        fd.write(comp)


def get_compressed_size(data, compression, config=None):
    """
    Returns the number of bytes required when the given data is
    compressed.

    Parameters
    ----------
    See `compress()`.

    Returns
    -------
    nbytes : int
        The size of the compressed data

    """

    class _ByteCountingFile:
        def __init__(self):
            self.count = 0

        def write(self, data):
            self.count += len(data)

    bcf = _ByteCountingFile()
    compress(bcf, data, compression, config=config)

    return bcf.count



import random

import pytest

import asdf
from asdf import IntegerType
from asdf.testing.helpers import roundtrip_object

# Make sure tests are deterministic
random.seed(0)


@pytest.mark.parametrize("sign", ["+", "-"])
@pytest.mark.parametrize(
    "value",
    [
        random.getrandbits(64),
        random.getrandbits(65),
        random.getrandbits(100),
        random.getrandbits(128),
        random.getrandbits(129),
        random.getrandbits(200),
    ],
)
def test_integer_value(value, sign):
    if sign == "-":
        value = -value

    integer = IntegerType(value)
    assert integer == roundtrip_object(integer)


@pytest.mark.parametrize("inline", [False, True])
def test_integer_storage(tmp_path, inline):
    tmpfile = str(tmp_path / "integer.asdf")

    kwargs = {}
    if inline:
        kwargs["storage_type"] = "inline"

    random.seed(0)
    value = random.getrandbits(1000)
    tree = {"integer": IntegerType(value, **kwargs)}

    with asdf.AsdfFile(tree) as af:
        af.write_to(tmpfile)

    tree = asdf.util.load_yaml(tmpfile, tagged=True)
    if inline:
        assert "source" not in tree["integer"]["words"]
        assert "data" in tree["integer"]["words"]
    else:
        assert "source" in tree["integer"]["words"]
        assert "data" not in tree["integer"]["words"]

    assert "string" in tree["integer"]
    assert tree["integer"]["string"] == str(value)


def test_integer_conversion():
    random.seed(0)
    value = random.getrandbits(1000)

    integer = asdf.IntegerType(value)
    assert integer == value
    assert int(integer) == int(value)
    assert float(integer) == float(value)



from asdf.extension import Converter


class ConstantConverter(Converter):
    tags = ["tag:stsci.edu:asdf/core/constant-1.0.0"]
    types = ["asdf.tags.core.constant.Constant"]

    def to_yaml_tree(self, obj, tag, ctx):
        return obj.value

    def from_yaml_tree(self, node, tag, ctx):
        from asdf.tags.core import Constant

        return Constant(node)



"""
This provides abstractions around a number of different file and
stream types available to Python so that they are always used in the
most efficient way.

The classes in this module should not be instantiated directly, but
instead, one should use the factory function `get_file`.
"""

import io
import mmap
import os
import pathlib
import re
import sys
from os import SEEK_CUR, SEEK_END, SEEK_SET
from urllib.request import url2pathname

import numpy as np

from . import util
from ._extern import atomicfile
from .exceptions import DelimiterNotFoundError
from .util import _patched_urllib_parse

__all__ = ["get_file", "get_uri", "relative_uri", "resolve_uri"]


_FILE_PERMISSIONS_DEFAULT_UMASK = 0o22
_FILE_PERMISSIONS_ALL = 0o777
_FILE_PERMISSIONS_NO_EXECUTE = 0o666

_local_file_schemes = ["", "file"]
if sys.platform.startswith("win"):  # pragma: no cover
    import string

    _local_file_schemes.extend(string.ascii_letters)


def _check_bytes(fd, mode):
    """
    Checks whether a given file-like object is opened in binary mode.
    """
    # On Python 3, doing fd.read(0) on an HTTPResponse object causes
    # it to not be able to read any further, so we do this different
    # kind of check, which, unfortunately, is not as robust.
    if isinstance(fd, io.IOBase):
        if isinstance(fd, io.TextIOBase):
            return False
        return True

    if "r" in mode:
        x = fd.read(0)
        if not isinstance(x, bytes):
            return False
    elif "w" in mode:
        try:
            fd.write(b"")
        except TypeError:
            return False

    return True


def resolve_uri(base, uri):
    """
    Resolve a URI against a base URI.
    """
    if base is None:
        base = ""
    resolved = _patched_urllib_parse.urljoin(base, uri)
    parsed = _patched_urllib_parse.urlparse(resolved)
    if parsed.path != "" and not parsed.path.startswith("/"):
        msg = "Resolved to relative URL"
        raise ValueError(msg)
    return resolved


def relative_uri(source, target):
    """
    Make a relative URI from source to target.
    """
    su = _patched_urllib_parse.urlparse(source)
    tu = _patched_urllib_parse.urlparse(target)
    extra = list(tu[3:])
    relative = None
    if tu[0] == "" and tu[1] == "":
        if tu[2] == su[2]:
            relative = ""
        elif not tu[2].startswith("/"):
            relative = tu[2]
    elif su[0:2] != tu[0:2]:
        return target

    if relative is None:
        relative = "" if tu[2] == su[2] else os.path.relpath(tu[2], os.path.dirname(su[2]))
    if relative == ".":
        relative = ""

    return _patched_urllib_parse.urlunparse(["", "", relative, *extra])


class _TruncatedReader:
    """
    Reads until a given delimiter is found.  Only works with
    RandomAccessFile and InputStream, though as this is a private
    class, this is not explicitly enforced.
    """

    def __init__(
        self,
        fd,
        delimiter,
        readahead_bytes,
        delimiter_name=None,
        include=False,
        initial_content=b"",
        exception=True,
    ):
        self._fd = fd
        self._delimiter = delimiter
        self._readahead_bytes = readahead_bytes
        if delimiter_name is None:
            delimiter_name = delimiter
        self._delimiter_name = delimiter_name
        self._include = include
        self._initial_content = initial_content
        self._trailing_content = b""
        self._exception = exception
        self._past_end = False

    def read(self, nbytes=None):
        if self._past_end:
            content = self._trailing_content[:nbytes]
            if nbytes is None:
                self._trailing_content = b""
            else:
                self._trailing_content = self._trailing_content[nbytes:]

            return content

        if nbytes is None:
            content = self._fd.peek()
        elif nbytes > len(self._initial_content):
            content = self._fd.peek(nbytes - len(self._initial_content) + self._readahead_bytes)
        else:
            content = self._initial_content[:nbytes]
            self._initial_content = self._initial_content[nbytes:]

            return content

        if content == b"":
            if self._exception:
                msg = f"{self._delimiter_name} not found"
                raise DelimiterNotFoundError(msg)

            self._past_end = True

            return content

        index = re.search(self._delimiter, content)
        if index is not None:
            index = index.end() if self._include else index.start()

            content = content[:index]
            self._past_end = True

        elif not (nbytes is None and self._exception):
            if nbytes:
                content = content[: nbytes - len(self._initial_content)]
        else:
            # Read the whole file and didn't find the delimiter
            msg = f"{self._delimiter_name} not found"
            raise DelimiterNotFoundError(msg)

        self._fd.fast_forward(len(content))

        if self._initial_content:
            content = self._initial_content + content
            self._initial_content = b""

        if self._past_end and nbytes:
            self._trailing_content = content[nbytes:]
            content = content[:nbytes]

        return content


class GenericFile(metaclass=util._InheritDocstrings):
    """
    Base class for an abstraction layer around a number of different
    file-like types.  Each of its subclasses handles a particular kind
    of file in the most efficient way possible.

    This class should not be instantiated directly, but instead the
    factory function `get_file` should be used to get the correct
    subclass for the given file-like object.
    """

    def __init__(self, fd, mode, close=False, uri=None):
        """
        Parameters
        ----------
        fd : file-like object
            The particular kind of file-like object must match the
            subclass of `GenericFile` being instantiated.

        mode : str
            Must be ``"r"`` (read), ``"w"`` (write), or ``"rw"``
            (read/write).

        close : bool, optional
            When ``True``, close the given `fd` in the ``__exit__``
            method, i.e. at the end of the with block.  Should be set
            to ``True`` when this object "owns" the file object.
            Default: ``False``.

        uri : str, optional
            The file path or URI used to open the file.  This is used
            to resolve relative URIs when the file refers to external
            sources.
        """
        if not _check_bytes(fd, mode):
            msg = "File-like object must be opened in binary mode."
            raise ValueError(msg)

        # can't import at the top level due to circular import
        from .config import get_config

        self._asdf_get_config = get_config

        self._fd = fd
        self._mode = mode
        self._close = close
        self._uri = uri

        self.block_size = get_config().io_block_size

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        if self._close:
            if hasattr(self._fd, "__exit__"):
                self._fd.__exit__(type_, value, traceback)
            else:
                self._fd.close()

    @property
    def block_size(self):
        return self._blksize

    @block_size.setter
    def block_size(self, block_size):
        if block_size == -1:
            try:
                block_size = os.fstat(self._fd.fileno()).st_blksize
            except Exception:
                block_size = io.DEFAULT_BUFFER_SIZE

        if block_size <= 0:
            msg = f"block_size ({block_size}) must be > 0"
            raise ValueError(msg)

        self._blksize = block_size

    @property
    def mode(self):
        """
        The mode of the file.  Will be ``'r'``, ``'w'`` or ``'rw'``.
        """
        return self._mode

    @property
    def uri(self):
        """
        The base uri of the file.
        """
        return self._uri

    def read(self, size=-1):
        """
        Read at most size bytes from the file (less if the read hits
        EOF before obtaining size bytes). If the size argument is
        negative or omitted, read all data until EOF is reached. The
        bytes are returned as a `bytes` object. An empty `bytes`
        object is returned when EOF is encountered immediately.

        Only available if `readable` returns `True`.
        """
        # On Python 3, reading 0 bytes from a socket causes it to stop
        # working, so avoid doing that at all costs.
        if size == 0:
            return b""
        return self._fd.read(size)

    def read_block(self):
        """
        Read a "block" from the file.  For real filesystem files, the
        block is the size of a native filesystem block.
        """
        return self.read(self._blksize)

    def read_blocks(self, size):
        """
        Read ``size`` bytes of data from the file, one block at a
        time.  The result is a generator where each value is a bytes
        object.
        """
        for i in range(0, size, self._blksize):
            thissize = min(self._blksize, size - i)
            yield self.read(thissize)

    def write(self, content):
        self._fd.write(content)

    write.__doc__ = """
    Write a string to the file. There is no return value. Due to
    buffering, the string may not actually show up in the file
    until the flush() or close() method is called.

    Only available if `writable` returns `True`.
    """

    def write_array(self, array):
        """
        Write array content to the file.  Array must be 1D contiguous
        so that this method can avoid making assumptions about the
        intended memory layout.  Endianness is preserved.

        Parameters
        ----------
        array : np.ndarray
            Must be 1D contiguous.
        """
        if len(array.shape) != 1 or not array.flags.contiguous:
            msg = "Requires 1D contiguous array."
            raise ValueError(msg)

        self.write(array.data)

    def peek(self, size=-1):
        """
        Read bytes of the file without consuming them.  This method
        must be implemented by all GenericFile implementations that
        provide ASDF input (those that aren't seekable should use a
        buffer to store peeked bytes).

        Parameters
        ----------
        size : int
            Number of bytes to peek, or -1 to peek all remaining bytes.
        """
        if self.seekable():
            cursor = self.tell()
            content = self.read(size)
            self.seek(cursor, SEEK_SET)

            return content

        msg = "Non-seekable file"
        raise RuntimeError(msg)

    def seek(self, offset, whence=0):
        """
        Set the file's current position.  Only available if `seekable`
        returns `True`.

        Parameters
        ----------
        offset : integer
            Offset, in bytes.

        whence : integer, optional
            The `whence` argument is optional and defaults to
            SEEK_SET or 0 (absolute file positioning); other values
            are SEEK_CUR or 1 (seek relative to the current
            position) and SEEK_END or 2 (seek relative to the
            file`s end).
        """
        result = self._fd.seek(offset, whence)
        return result

    def tell(self):
        """
        Return the file's current position, in bytes.  Only available
        in `seekable` returns `True`.
        """
        return self._fd.tell()

    def flush(self):
        """
        Flush the internal buffer.
        """
        self._fd.flush()

    def close(self):
        """
        Close the file.  The underlying file-object will only be
        closed if ``close=True`` was passed to the constructor.
        """
        if self._close:
            self._fd.close()
            if hasattr(self, "_fix_permissions"):
                self._fix_permissions()

    def truncate(self, size=None):
        """
        Truncate the file to the given size.
        """
        raise NotImplementedError

    def writable(self):
        """
        Returns `True` if the file can be written to.
        """
        return "w" in self.mode

    def readable(self):
        """
        Returns `True` if the file can be read from.
        """
        return "r" in self.mode

    def seekable(self):
        """
        Returns `True` if the file supports random access (`seek` and
        `tell`).
        """
        return False

    def can_memmap(self):
        """
        Returns `True` if the file supports memmapping.
        """
        return False

    def is_closed(self):
        """
        Returns `True` if the underlying file object is closed.
        """
        return self._fd.closed

    def read_until(
        self,
        delimiter,
        readahead_bytes,
        delimiter_name=None,
        include=True,
        initial_content=b"",
        exception=True,
    ):
        """
        Reads until a match for a given regular expression is found.

        Parameters
        ----------
        delimiter : str
            A regular expression.

        readahead_bytes : int
            The number of bytes to read ahead to make sure the
            delimiter isn't on a block boundary.

        delimiter_name : str, optional
            The name of the delimiter.  Used in error messages if the
            delimiter is not found.  If not provided, the raw content
            of `delimiter` will be used.

        include : bool, optional
            When ``True``, include the delimiter in the result.

        initial_content : bytes, optional
            Additional content to include at the beginning of the
            first read.

        exception : bool, optional
            If ``True`` (default), raise an exception if the end
            marker isn't found.

        Returns
        -------
        content : bytes
            The content from the current position in the file, up to
            the delimiter.  Includes the delimiter if `include` is
            ``True``.

        Raises
        ------
        DelimiterNotFoundError :
            If the delimiter is not found before the end of the file.
        """
        buff = io.BytesIO()
        reader = self.reader_until(
            delimiter,
            readahead_bytes,
            delimiter_name=delimiter_name,
            include=include,
            initial_content=initial_content,
            exception=exception,
        )
        while True:
            content = reader.read(self.block_size)
            buff.write(content)
            if len(content) < self.block_size:
                break
        return buff.getvalue()

    def reader_until(
        self,
        delimiter,
        readahead_bytes,
        delimiter_name=None,
        include=True,
        initial_content=b"",
        exception=True,
    ):
        """
        Returns a readable file-like object that treats the given
        delimiter as the end-of-file.

        Parameters
        ----------
        delimiter : str
            A regular expression.

        readahead_bytes : int
            The number of bytes to read ahead to make sure the
            delimiter isn't on a block boundary.

        delimiter_name : str, optional
            The name of the delimiter.  Used in error messages if the
            delimiter is not found.  If not provided, the raw content
            of `delimiter` will be used.

        include : bool, optional
            When ``True``, include the delimiter in the result.

        initial_content : bytes, optional
            Additional content to include at the beginning of the
            first read.

        exception : bool, optional
            If ``True`` (default), raise an exception if the end
            marker isn't found.

        Raises
        ------
        DelimiterNotFoundError :
            If the delimiter is not found before the end of the file.
        """
        raise NotImplementedError

    def seek_until(
        self,
        delimiter,
        readahead_bytes,
        delimiter_name=None,
        include=True,
        initial_content=b"",
        exception=True,
    ):
        """
        Seeks in the file until a match for a given regular expression
        is found.  This is similar to ``read_until``, except the
        intervening content is not retained.

        Parameters
        ----------
        delimiter : str
            A regular expression.

        readahead_bytes : int
            The number of bytes to read ahead to make sure the
            delimiter isn't on a block boundary.

        delimiter_name : str, optional
            The name of the delimiter.  Used in error messages if the
            delimiter is not found.  If not provided, the raw content
            of `delimiter` will be used.

        include : bool, optional
            When ``True``, include the delimiter in the result.

        initial_content : bytes, optional
            Additional content to include at the beginning of the
            first read.

        exception : bool, optional
            If ``True`` (default), raise an exception if the end
            marker isn't found.

        Returns
        -------
        bool
            ``True`` if the delimiter was found.

        Raises
        ------
        DelimiterNotFoundError :
            If ``exception`` is enabled and the delimiter is not found
            before the end of the file.
        """
        reader = self.reader_until(
            delimiter,
            readahead_bytes,
            delimiter_name=delimiter_name,
            include=include,
            initial_content=initial_content,
            exception=True,
        )
        try:
            while reader.read(self.block_size) != b"":
                pass
        except DelimiterNotFoundError:
            if exception:
                raise

            return False

        return True

    def fast_forward(self, size):
        """
        Move the file position forward by `size`.
        """
        raise NotImplementedError

    def clear(self, nbytes):
        """
        Write nbytes of zeros.
        """
        blank_data = b"\0" * self.block_size
        for i in range(0, nbytes, self.block_size):
            length = min(nbytes - i, self.block_size)
            self.write(blank_data[:length])

    def memmap_array(self, offset, size):
        """
        Memmap a chunk of the file into a `np.memmap` object.

        Parameters
        ----------
        offset : integer
            The offset, in bytes, in the file.

        size : integer
            The size of the data to memmap.

        Returns
        -------
        array : np.memmap
        """
        msg = f"memmapping is not implemented for {self.__class__.__name__}"
        raise NotImplementedError(msg)

    def close_memmap(self):
        """
        Close the memmapped file (if one was mapped with memmap_array)
        """
        msg = f"memmapping is not implemented for {self.__class__.__name__}"
        raise NotImplementedError(msg)

    def flush_memmap(self):
        """
        Flush any pending writes to the memmapped file (if one was mapped with
        memmap_array)
        """
        msg = f"memmapping is not implemented for {self.__class__.__name__}"
        raise NotImplementedError(msg)

    def read_into_array(self, size):
        """
        Read a chunk of the file into a uint8 array.

        Parameters
        ----------
        size : integer
            The size of the data.

        Returns
        -------
        array : np.memmap
        """
        buff = self.read(size)
        return np.frombuffer(buff, np.uint8, size, 0)


class GenericWrapper:
    """
    A wrapper around a `GenericFile` object so that closing only
    happens in the very outer layer.
    """

    def __init__(self, fd):
        self._fd = fd

    def __enter__(self):
        return self

    def __exit__(self, type_, value, traceback):
        pass

    def __getattr__(self, attr):
        return getattr(self._fd, attr)


class RandomAccessFile(GenericFile):
    """
    The base class of file types that support random access.
    """

    def seekable(self):
        return True

    def reader_until(
        self,
        delimiter,
        readahead_bytes,
        delimiter_name=None,
        include=True,
        initial_content=b"",
        exception=True,
    ):
        return _TruncatedReader(
            self,
            delimiter,
            readahead_bytes,
            delimiter_name=delimiter_name,
            include=include,
            initial_content=initial_content,
            exception=exception,
        )

    def fast_forward(self, size):
        if size < 0:
            self.seek(0, SEEK_END)
        self.seek(size, SEEK_CUR)

    def truncate(self, size=None):
        if size is None:
            self._fd.truncate()
        else:
            self._fd.truncate(size)
            self.seek(size, SEEK_SET)


class RealFile(RandomAccessFile):
    """
    Handles "real" files on a filesystem.
    """

    def __init__(self, fd, mode, close=False, uri=None):
        super().__init__(fd, mode, close=close, uri=uri)

        if uri is None and hasattr(fd, "name") and isinstance(fd.name, str):
            self._uri = pathlib.Path(fd.name).expanduser().absolute().as_uri()

    def write_array(self, arr):
        if isinstance(arr, np.memmap) and getattr(arr, "fd", None) is self:
            arr.flush()
            self.fast_forward(len(arr.data))
        else:
            if len(arr.shape) != 1 or not arr.flags.contiguous:
                msg = "Requires 1D contiguous array."
                raise ValueError(msg)

            self._fd.write(arr.data)

    def can_memmap(self):
        return True

    def memmap_array(self, offset, size):
        if not hasattr(self, "_mmap"):
            loc = self._fd.tell()
            acc = mmap.ACCESS_WRITE if "w" in self._mode else mmap.ACCESS_READ
            self._fd.seek(0, 2)
            nbytes = self._fd.tell()
            self._mmap = mmap.mmap(self._fd.fileno(), nbytes, access=acc)
            # on windows mmap seeks to the start of the file so return the file
            # pointer to this previous location
            self._fd.seek(loc, 0)
        return np.ndarray.__new__(np.memmap, shape=size, offset=offset, dtype="uint8", buffer=self._mmap)

    def close_memmap(self):
        if hasattr(self, "_mmap"):
            # we no longer close the _mmap here. This does mean that views of arrays
            # that are backed by _mmap will keep the _mmap alive (and open). This is
            # the cost of avoiding segfaults as np.memmap does not check if mmap is
            # closed.
            del self._mmap

    def flush_memmap(self):
        if hasattr(self, "_mmap"):
            self._mmap.flush()

    def read_into_array(self, size):
        return np.fromfile(self._fd, dtype=np.uint8, count=size)

    def _fix_permissions(self):
        """
        atomicfile internally uses tempfile.NamedTemporaryFile
        which uses tempfile.mkstemp which makes a file that is

        "readable and writable only by the creating user ID."

        this creates files with mode 0o600 regardless of umask
        Rather than modify atomicfile, this will use
        the umask to determine the file permissions and modify
        the resulting file permission bits
        """
        if isinstance(self._fd, atomicfile._AtomicWFile):
            fn = self._fd._filename
            if not os.path.exists(fn):
                return
            # there is no way to read the umask without setting it
            # so set it to the typical default 0o22
            umask = os.umask(_FILE_PERMISSIONS_DEFAULT_UMASK)
            if umask != _FILE_PERMISSIONS_DEFAULT_UMASK:
                # restore the read value if it differs from the default
                os.umask(umask)
            permissions = _FILE_PERMISSIONS_ALL if os.path.isdir(fn) else _FILE_PERMISSIONS_NO_EXECUTE
            os.chmod(self._fd._filename, permissions & ~umask)

    def __exit__(self, type_, value, traceback):
        super().__exit__(type_, value, traceback)
        if self._close:
            self._fix_permissions()

    def close(self):
        self.flush_memmap()
        super().close()
        self.close_memmap()
        if self._close:
            self._fix_permissions()

    def truncate(self, size=None):
        # windows supports truncating as long as the file not opened
        # more than once. So this must be called after closing all
        # memmaps
        if sys.platform.startswith("win") and hasattr(self, "_mmap"):
            self._mmap.close()
            self.close_memmap()
        super().truncate(size=size)


class MemoryIO(RandomAccessFile):
    """
    Handles random-access memory buffers, mainly `io.BytesIO` and
    `StringIO.StringIO`.
    """

    def __init__(self, fd, mode, uri=None):
        super().__init__(fd, mode, uri=uri)

    def read_into_array(self, size):
        buf = self._fd.getvalue()
        offset = self._fd.tell()
        result = np.frombuffer(buf, np.uint8, size, offset)
        # Copy the buffer so the original memory can be released.
        result = result.copy()
        self.seek(size, SEEK_CUR)
        return result


class InputStream(GenericFile):
    """
    Handles an input stream, such as stdin.
    """

    def __init__(self, fd, mode="r", close=False, uri=None):
        super().__init__(fd, mode, close=close, uri=uri)
        self._fd = fd
        self._buffer = b""

    def peek(self, size=-1):
        if size < 0:
            self._buffer += self._fd.read()
        else:
            len_buffer = len(self._buffer)
            if len_buffer < size:
                self._buffer += self._fd.read(size - len_buffer)
        return self._buffer

    def read(self, size=-1):
        # On Python 3, reading 0 bytes from a socket causes it to stop
        # working, so avoid doing that at all costs.
        if size == 0:
            return b""

        len_buffer = len(self._buffer)
        if len_buffer == 0:
            return self._fd.read(size)

        if size < 0:
            self._buffer += self._fd.read()
            buffer = self._buffer
            self._buffer = b""

            return buffer

        if len_buffer < size:
            if len_buffer < size:
                self._buffer += self._fd.read(size - len(self._buffer))

            buffer = self._buffer
            self._buffer = b""

            return buffer

        buffer = self._buffer[:size]
        self._buffer = self._buffer[size:]

        return buffer

    def reader_until(
        self,
        delimiter,
        readahead_bytes,
        delimiter_name=None,
        include=True,
        initial_content=b"",
        exception=True,
    ):
        return _TruncatedReader(
            self,
            delimiter,
            readahead_bytes,
            delimiter_name=delimiter_name,
            include=include,
            initial_content=initial_content,
            exception=exception,
        )

    def fast_forward(self, size):
        if size >= 0 and len(self.read(size)) != size:
            msg = "Read past end of file"
            raise OSError(msg)

    def read_into_array(self, size):
        try:
            # See if Numpy can handle this as a real file first...
            return np.fromfile(self._fd, np.uint8, size)
        except (OSError, AttributeError):
            # Else, fall back to reading into memory and then
            # returning the Numpy array.
            data = self.read(size)
            # We need to copy the array, so it is writable
            result = np.frombuffer(data, np.uint8, size)
            # When creating an array from a buffer, it is read-only.
            # If we need a read/write array, we have to copy it.
            if "w" in self._mode:
                result = result.copy()
            return result


class OutputStream(GenericFile):
    """
    Handles an output stream, such as stdout.
    """

    def __init__(self, fd, close=False, uri=None):
        super().__init__(fd, "w", close=close, uri=uri)
        self._fd = fd

    def fast_forward(self, size):
        if size < 0:
            return
        self.clear(size)


def get_uri(file_obj):
    """
    Returns the uri of the given file object

    Parameters
    ----------
    uri : object
    """
    if isinstance(file_obj, str):
        return file_obj
    if isinstance(file_obj, GenericFile):
        return file_obj.uri

    # A catch-all for types from Python's io module that have names
    return getattr(file_obj, "name", "")


def get_file(init, mode="r", uri=None, close=False):
    """
    Returns a `GenericFile` instance suitable for wrapping the given
    object `init`.

    If passed an already open file-like object, it must be opened for
    reading/writing in binary mode.  It is the caller's responsibility
    to close it.

    Parameters
    ----------
    init : object
        `init` may be:

        - A `bytes` or `unicode` file path or ``file:`` or ``http:``
          url.

        - A Python 2 `file` object.

        - An `io.IOBase` object (the default file object on Python 3).

        - A ducktyped object that looks like a file object.  If `mode`
          is ``"r"``, it must have a ``read`` method.  If `mode` is
          ``"w"``, it must have a ``write`` method.  If `mode` is
          ``"rw"`` it must have the ``read``, ``write``, ``tell`` and
          ``seek`` methods.

        - A `GenericFile` instance, in which case it is wrapped in a
          `GenericWrapper` instance, so that the file is closed when
          only when the final layer is unwrapped.

    mode : str
        Must be one of ``"r"``, ``"w"`` or ``"rw"``.

    uri : str
        Sets the base URI of the file object.  This will be used to
        resolve any relative URIs contained in the file.  This is
        redundant if `init` is a `bytes` or `unicode` object (since it
        will be the uri), and it may be determined automatically if
        `init` refers to a regular filesystem file.  It is not required
        if URI resolution is not used in the file.

    close : bool
        If ``True``, closes the underlying file handle when this object is
        closed. Defaults to ``False``.

    Returns
    -------
    fd : GenericFile

    Raises
    ------
    ValueError, TypeError, IOError
    """
    if mode not in ("r", "w", "rw"):
        msg = "mode must be 'r', 'w' or 'rw'"
        raise ValueError(msg)

    if init in (sys.__stdout__, sys.__stdin__, sys.__stderr__):
        init = os.fdopen(init.fileno(), init.mode + "b")

    if isinstance(init, (GenericFile, GenericWrapper)):
        if mode not in init.mode:
            msg = f"File is opened as '{init.mode}', but '{mode}' was requested"
            raise ValueError(msg)

        return GenericWrapper(init)

    if isinstance(init, (str, pathlib.Path)):
        parsed = _patched_urllib_parse.urlparse(str(init))

        realmode = "r+b" if mode == "rw" else mode + "b"

        if parsed.scheme in _local_file_schemes:
            # if paths have an extra leading '/' urlparse will
            # parse them even though they violate rfc8089. This will
            # lead to errors or writing files to unexpected locations
            # on non-windows systems
            if not sys.platform.startswith("win") and parsed.scheme == "" and parsed.netloc != "":
                msg = f"Invalid path {init}"
                raise ValueError(msg)

            # Windows paths are not URIs, and so they should not be parsed as
            # such. Otherwise, the drive component of the path can get lost.
            # This is not an ideal solution, but we can't use pathlib here
            # because it doesn't handle URIs properly.
            realpath = (
                str(init)
                if sys.platform.startswith("win") and parsed.scheme in string.ascii_letters
                else url2pathname(parsed.path)
            )
            try:
                fd = atomicfile.atomic_open(realpath, realmode) if mode == "w" else open(realpath, realmode)

                fd = fd.__enter__()
            except FileNotFoundError as e:
                # atomic_open will create an Exception with an odd looking path
                # overwrite the error message to make it more informative
                e.filename = realpath
                raise e

            return RealFile(fd, mode, close=True, uri=uri)

        # this is not a local file, import fsspec (if available)
        try:
            import fsspec
        except ImportError:
            fsspec = None

        if fsspec:
            fd = fsspec.open(init, realmode)
            try:
                fd = fd.__enter__()
            except NotImplementedError as err:
                msg = f"Unable to open {init} with mode {mode}"
                raise ValueError(msg) from err
            return get_file(fd, uri=uri or init, close=True)

    if isinstance(init, io.BytesIO):
        return MemoryIO(init, mode, uri=uri)

    if isinstance(init, io.StringIO):
        msg = "io.StringIO objects are not supported.  Use io.BytesIO instead."
        raise TypeError(msg)

    if isinstance(init, io.IOBase):
        if ("r" in mode and not init.readable()) or ("w" in mode and not init.writable()):
            msg = f"File is opened as '{init.mode}', but '{mode}' was requested"
            raise ValueError(msg)

        if init.seekable():
            init2 = init.raw if hasattr(init, "raw") else init

            if hasattr(init2, "getvalue"):
                result = MemoryIO(init2, mode, uri=uri)
            else:
                # can we call 'fileno'? if not, we can't memmap so don't
                # make a RealFile
                try:
                    init2.fileno()
                    result = RealFile(init2, mode, uri=uri, close=close)
                except io.UnsupportedOperation:
                    result = RandomAccessFile(init2, mode, uri=uri, close=close)

            result._secondary_fd = init
            return result

        if mode == "w":
            return OutputStream(init, uri=uri, close=close)

        if mode == "r":
            return InputStream(init, mode, uri=uri, close=close)

        msg = f"File '{init}' could not be opened in 'rw' mode"
        raise ValueError(msg)

    if mode == "w" and (hasattr(init, "write") and hasattr(init, "seek") and hasattr(init, "tell")):
        return MemoryIO(init, mode, uri=uri)

    if mode == "r" and (hasattr(init, "read") and hasattr(init, "seek") and hasattr(init, "tell")):
        return MemoryIO(init, mode, uri=uri)

    if mode == "rw" and (
        hasattr(init, "read") and hasattr(init, "write") and hasattr(init, "seek") and hasattr(init, "tell")
    ):
        return MemoryIO(init, mode, uri=uri)

    if mode == "w" and hasattr(init, "write"):
        return OutputStream(init, uri=uri, close=close)

    if mode == "r" and hasattr(init, "read"):
        return InputStream(init, mode, uri=uri, close=close)

    msg = f"Can't handle '{init}' as a file for mode '{mode}'"
    raise ValueError(msg)
